package ViewAbstraction

import scala.collection.mutable.ArrayBuffer

/** A checker for the view abstraction algorithm, applied to system.
  * @param aShapes the shapes of abstractions.
  * @param cShapes the shapes of concretizations. */
class Checker(system: System, aShapes: List[Array[Int]], cShapes: List[Array[Int]]){
  // is there a three-way synchronisation between two components and a server?
  private val threeWaySync = system.threeWaySyncFound

  /** The abstract views. */
  var sysAbsViews: SystemViewSet = null

  /** The concretizations. */
  val sysConcViews = SystemViewSet()

  /** The first view extender. */
  var nve1: NewViewExtender = null

  /** The second view extender. */
  var nve2: NewViewExtender = null

  /** Run the checker. 
    * @param bound the number of plys to explore (with negative values meaning 
    * effectively infinite).  */
  def apply(bound: Int = -1, verbose: Boolean = false)  = {
    // Note: in comments below, k refers to the size of abstractions, the sum
    // of elements of aShapes; l refers to the size of concretizations, the
    // sum of elements of cShapes.  threeWaySync is true iff l = k+2.

    sysAbsViews = system.initViews(aShapes)
    val initViews = new ArrayBuffer[SystemView] 
    initViews ++= sysAbsViews.toList
    println("init:\n" + sysAbsViews.mkString("\n"))
    // All shapes formed by adding one component to a shape from aShapes
    var midShapes = List[Array[Int]]()
    if(threeWaySync){
      for(sh <- aShapes; i <- 0 until numTypes){
        val newShape = sh.clone; newShape(i) += 1
        if(midShapes.forall(sh1 => !sh1.sameElements(newShape)))
          midShapes ::= newShape
      }
      println("midShapes = "+midShapes.map(_.mkString("<", ",", ">")))
    }
    // View extender to extend from aShapes to midShapes
    nve1 =
      if(!threeWaySync)
        new NewViewExtender(sysConcViews, aShapes, cShapes)
      else{
        assert(numTypes <= 2)
        new NewViewExtender(SystemViewSet(), aShapes, midShapes)
      }
    // View extender to extend from midShapes to cShapes
    nve2 = 
      if(!threeWaySync) null
      else new NewViewExtender(sysConcViews, midShapes, cShapes)
    // The gamma function
    @inline def gamma(newViews: Array[SystemView]) = 
      if(!threeWaySync) nve1.gamma(newViews)
      else{
        val newMidViews = nve1.gamma(newViews) // of length k+1
        nve2.gamma(newMidViews) // of length k+2
      }

    // Views seen for the first time in the last iteration
    var newViews: Array[SystemView] = null
    newViews = gamma(initViews.toArray) // IMPROVE
    //system.checkTypeSizes(newViews)

    if(true || verbose)
      println("init sysConcViews:\n" + sysConcViews.mkString("\n"))
    // println("state count = "+State.stateCount)

    // Main loop
    var done = false; var ply = 0
    while(!done && ply != bound){
      ply += 1; println("\nSTEP "+ply)
      val oldSize = sysConcViews.size
      println("#concretizations = "+printInt(oldSize))
      println("#new concretizations = "+printInt(newViews.size))
      val (cPost, errors, deadlocks) = 
        system.post(newViews, sysConcViews, verbose)
      println("After post: "+printInt(cPost.size))
      if(errors.nonEmpty){
        println("error event possible from:\n" + errors.take(30).mkString("\n"))
        println(errors.length+" error states")
        val debugger =
          new Debugger(system, threeWaySync, aShapes, sysAbsViews, 
                       initViews.toList, false)
        debugger(errors.head)
        done = true
      }
      else if(deadlocks.nonEmpty){
        println("deadlock in:\n" + deadlocks.take(30).mkString("\n"))
        println(deadlocks.length+" deadlocked states")
        val debugger =
          new Debugger(system, threeWaySync, aShapes, sysAbsViews, 
                       initViews.toList, true)
        debugger(deadlocks.head)
        done = true
      }
      else{
        val newAbs = system.alpha(aShapes, cPost, sysAbsViews, ply)
        println("#new abstract views = "+printInt(newAbs.size))
        if(verbose) println(newAbs.mkString("\n"))
        newViews = gamma(newAbs) // : Array[SystemView]
        newViews = newViews ++ cPost // IMPROVE, avoid this
        //system.checkTypeSizes(newViews)
        if(verbose) println("sysConcViews "+sysConcViews.size+":")
        done = sysConcViews.size == oldSize
        println("abstract states = "+printInt(sysAbsViews.size))
        if(true || verbose || done || ply == bound){
          println("state count = "+printInt(State.stateCount))
          println("server state count = "+ServerStates.count)
          println("server transStoreSize = "+system.servers.transStoreSize)
          println("nve1.size = "+printInt(nve1.size)+
                    "; stateCount = "+printInt(nve1.stateCount))
          if(nve2 != null)
            println("nve2.size = "+printInt(nve2.size)+
                      "; stateCount = "+printInt(nve2.stateCount))
        }
      }
    } // end of while(!done)
    // println(sysConcViews.reportSizes.mkString(", "))
    println("Explored "+printInt(sysConcViews.size)+" system states.")
  }
}
