package ViewAbstraction

import uk.ac.ox.cs.fdr._
import scala.collection.JavaConverters._
import scala.collection.mutable.{Map,Set,ArrayBuffer}
// import ox.cads.util.Profiler

/** Model of the replicated components, possibly comprising multiple families
  * of processes.
  * 
  * @param processNames the process name (from the script) for each type of 
  * component process.
  * @param fdrTypeIds the representation of each type inside FDR. 
  * @param alphaNames the name of the alphabet for each type components process.
  */
class Components(
  fdrSession: FDRSession, transMapBuilder: FDRTransitionMap,
  fdrTypeIds: Array[Long], processNames: Array[String], alphaNames: Array[String]
){
  // Check lengths agree
  println("fdrTypeIds = "+fdrTypeIds.mkString(",")+
            "\nprocessNames = "+processNames.mkString(",")+
            "\nalphaNames = "+alphaNames.mkString(","))

  assert(fdrTypeIds.length == numFamilies &&  
           processNames.length == numFamilies &&
           alphaNames.length == numFamilies)

  /** list of identities, for each process type. */
  private val idTypes: Array[Array[String]] =
    familyTypeNames.map(fdrSession.getTypeValues(_).toArray)
  assert(idTypes.length == numFamilies)

  /** Number of distinct identities in the script for each indexing type
    * (i.e. excluding distinguished values), indexed by family numbers. */
  private val idSizes: Array[Int] = idTypes.map(_.length)
  // Note: the above ends up holding the same values as typeSizes, but
  // the latter is not yet set.

  /** Identities of components for each process type. */
  private val indices: Array[List[Identity]] =
    idSizes.map(idSize => (0 until idSize).toList)

  /** Representation of the transitions of a component from a state, as a pair
    * of parallel arrays, giving each possible event, and the list of
    * successor states, respectively. */
  type Trans = (ArrayBuffer[EventInt], ArrayBuffer[List[State]])

  /** The transitions of the components: Map from states to a pair of parallel
    * arrays, giving each possible event, and the list of successor states,
    * respectively. */
  private var transMap = Map[State, Trans]()

  /** The initial states of each type. */
  private val inits: Array[Array[State]] =
    Array.tabulate(numFamilies)(i => new Array[State](idSizes(i)))

  /** The representation of the event tau. */
  // @inline private val Tau = 1

  /** Bit map, indexed by events, showing whether each event is in the alphabet
    * of these components. */
  var alphaBitMap: Array[Boolean] = null

  /** Map that gives, for each event e, the identities of components that
    * synchronise on e. */
  private var eventMap: Array[List[ComponentProcessIdentity]] = null

  def getEventMap = eventMap

  /** Show a particular event. */
  private def showEvent(e: EventInt) = fdrSession.eventToString(e)

  // Set global variable: the type identifiers of the indexing types,
  // by component family.
  indexingTypes = fdrTypeIds.map(ti => transMapBuilder.fdrTypeToType(ti))
  println("indexingTypes = "+indexingTypes.mkString(", "))

  typeNames = 
    (0 until numTypes).map(i => 
      familyTypeNames(indexingTypes.indexOf(i))).toArray
  println("typeNames = "+typeNames.mkString(", "))

  /** Build transition system for components in transMap, and alphabets for each
    * component in alphas, initial states in inits. */
  private def init() = {
    println("Building component alphabets")
    initAlphas() // Initialise eventMap and alphaBitMap
    println("Building component transitions")
    initTransitions()

    // For the ith family, the number of identities should be the same as the 
    // number of values in the corresponding type.
    assert((0 until numFamilies).forall(i => 
             idSizes(i) == typeSizes(indexingTypes(i))),
           idSizes.mkString("(",",",")")+"; "+typeSizes.mkString("(",",",")"))

    // Set maxSubtypeSize and build eventBitMap (note the former is
    // necessary for the latter).  IMPROVE: at present, typeSizes is
    // set in initTransitions; can this be moved into initAlphas?
    maxSubtypeSize = idSizes.max
    println("maxSubtypeSize = "+maxSubtypeSize)
    assert(maxSubtypeSize > 0)
  }

  /** Initialise information relevant to the alphabets: eventMap and
    * alphaBitMap. */
  private def initAlphas() = { 
    // Build alphabets: alphas(pt)(j) gives the alphabet of the jth element
    // (identity idTypes(pt)(j)) from process type pt.
    val alphas: Array[Array[scala.collection.immutable.Set[EventInt]]] =
      Array.tabulate(numFamilies)( f =>
        new Array[scala.collection.immutable.Set[EventInt]](idSizes(f)))
    for(f <- 0 until numFamilies; i <- indices(f)){
      // String representing name of alphabet of this process
      val alphaSt = alphaNames(f)+"("+idTypes(f)(i)+")"
      // Corresponding list of event names
      val alphaList: List[String] = fdrSession.setStringToList(alphaSt)
      // Convert to EventInts, and store
      alphas(f)(i) = alphaList.map(fdrSession.eventToInt(_)).toSet
    }

    val allEvents = alphas.map(_.flatten).flatten.distinct.sorted
    // Build eventMap, mapping each event e to the list of identities of
    // processes that synchronise on e.
    eventMap = Array.tabulate[List[Parameter]](eventsSize)(e =>
      (for(f <- 0 until numFamilies;
           i <- indices(f); if (alphas(f)(i).contains(e)))
       yield (f,i)).toList
    )

    for(e <- 0 until eventsSize)
      assert(eventMap(e).length <= 2,
             "Event "+showEvent(e)+" synchronised on by "+
               eventMap(e).length+" components.")
    // Build alphaBitmap showing which events are in alphabet; IMPROVE:
    // inefficient
    alphaBitMap = Array.tabulate(eventsSize)(e => allEvents.contains(e))
  }

  /** Initialise model of the transition systems: inits and transMap. */
  private def initTransitions() = {
    // Build transitions
    for(pt <- 0 until numFamilies){
      println("Building transitions for type "+processNames(pt))
      val seen = Set[State]() // states seen so far
      val transMap0 = Map[State, List[(EventInt, State)]]()
      for(i <- indices(pt)){
        val t = idTypes(pt)(i); val id = fdrSession.symmetryValue(t)
        val machine = fdrSession.evalProc(processNames(pt)+"("+t+")")
        // Build transition map
        inits(pt)(i) = transMapBuilder.augmentTransMap(
          transMap0, seen, machine, pt, id, fdrTypeIds(pt))
      }
      // Now convert into more convenient form, copying into transMap, giving
      // a map from states to parallel ArrayBuffers, giving possible events
      // and next states.
      println("# component states = "+seen.size)
      for(s <- seen){
        // Transitions still to consider
        var transList: List[(EventInt, State)] = transMap0(s)
        // Result for this state.
        val transListEvent = new ArrayBuffer[EventInt]
        val transListNexts = new ArrayBuffer[List[State]]
        for(e <- 0 until numEvents){
          assert(transList.isEmpty || transList.head._1 >= e)
          val (matches, rest) = transList.span(_._1 == e)
          transList = rest
          if(matches.nonEmpty){
            transListEvent += e; transListNexts += matches.map(_._2)
          }
        }
        transMap(s) = (transListEvent, transListNexts)
      }
    } // end of for(pt <- ...)
  } // end of initTransitions

  // Do the initialisation. 
  init()

  // -------------------------------------------------------

  /** Representation of the transitions from a state. */ 
  class ComponentTransitions{
    /* Transitions are represented by parallel arrays of events (terminated with
     * Sentinel) and lists of next-states. */

    // Those transitions synchronising with no other component, and
    // synchronising with the server.
    val eventsServer = new ArrayBuffer[EventInt]
    val nextsServer = new ArrayBuffer[List[State]]

    // Those transitions synchronising with no other component, nor with the
    // server.
    val eventsSolo = new ArrayBuffer[EventInt]
    val nextsSolo = new ArrayBuffer[List[State]]

    /** Transitions synchronising with one other component:
      * transComponent(f)(i) represents transitions synchronised with
      * the component with process identity (f,i); except null is used
      * to repsesent no transitions (for compactness). */
    val transComponent = Array.tabulate(numFamilies)(f =>
      Array.fill[Trans](idSizes(f))(
        (new ArrayBuffer[EventInt], new ArrayBuffer[List[State]])
      ) )

    /** Transitions synchronising with server and one other component:
      * representation is as for transComponent. */
    val transServerComponent = Array.tabulate(numFamilies)(f =>
      Array.fill[Trans](idSizes(f))(
        (new ArrayBuffer[EventInt], new ArrayBuffer[List[State]])
      ) )

    /** Add sentinels, and replace empty sets of transitions in transComponent by
      * null. */
    def finish = {
      // Add sentinels
      eventsServer += Sentinel; eventsSolo += Sentinel
      for(f <- 0 until numFamilies; i <- 0 until idSizes(f)){
        val es = transComponent(f)(i)._1
        if(es.nonEmpty) es += Sentinel else transComponent(f)(i) = null
        val es1 = transServerComponent(f)(i)._1
        if(es1.nonEmpty) es1 += Sentinel else transServerComponent(f)(i) = null
      }
    }
    // IMPROVE: use Arrays instead of ArrayBuffers?  Compress more?
  } // end of class ComponentTransitions

  // -------------------------------------------------------

  /** Map giving the transitions from a state. */ 
  private val transComponentStore = Map[State, ComponentTransitions]()
  /* Note: the above is created (sequentially) during initialisation,
   * and then read during the search, so doesn't need to be
   * protected. */

  /** Build a map, categorising transitions from each state as a value of type
    * ComponentTransitions, and storing these. */
  def categoriseTrans(serverAlphaBitMap: Array[Boolean]) = {
    for((st, (events0, nexts0)) <- transMap.iterator){
      val trans1 = new ComponentTransitions
      var i = 0; var n = events0.length
      if(n > 0 && events0(0) == Tau){ 
        trans1.eventsSolo += Tau; trans1.nextsSolo += nexts0(0); i = 1
      }
      val me: Parameter = st.componentProcessIdentity 
      while(i < n){
        val e = events0(i); val nexts = nexts0(i)
        val cpts = eventMap(e); assert(cpts.contains(me))
        if(cpts.length == 1){
          if(serverAlphaBitMap(e)){
            trans1.eventsServer += e; trans1.nextsServer += nexts
          }
          else{ trans1.eventsSolo += Tau; trans1.nextsSolo += nexts }
        }
        else{ // sync with one other component
          assert(cpts.length == 2)
          val head = cpts.head
          // the other component synchronising on e
          val (f,i) = if(head == me) cpts.tail.head else head
          val (es,nextss) =
            if(serverAlphaBitMap(e)) trans1.transServerComponent(f)(i)
            else trans1.transComponent(f)(i)
          es += e; nextss += nexts
        }
        i += 1
      }
      trans1.finish
      transComponentStore += st -> trans1
    } // end of for
    transMap = null // for GC
    // eventMap = null // IMPROVE: do this; also the ref to this in Servers
  }

  // --------- Functions used during the main check ---------

  /** Initial views with shapes from aShapes. */
  def initViews(aShapes: List[Array[Int]]): List[Views.View] = {
    val iv = Views.alpha(aShapes, inits.flatten)
    // println("aShapes = "+aShapes.map(_.mkString("<", ",", ">")))
    // Check that for each family f, there are enough identities to produce an
    // initial view with all components in the same control state.
    for(f <- 0 until numFamilies){
      val maxf = aShapes.map(_(f)).max // max cpts of f in any shape
      // println(s"$f max number = $maxf")
      // Test if v satisfies the condition
      def ok(v: Views.View) = {
        val vf = v.filter(_.family == f)
        vf.length == maxf && vf.map(_.cs).distinct.length == 1
      }
      if(!iv.exists(ok)){
        println("Not enough identities of family "+familyTypeNames(f)+
                  " to produce initial view with all components in default state.");
        sys.exit
      }
    }
    iv
  }

  /** Get the transitions of the component in state st. */
  def getTransComponent(st: State): ComponentTransitions =
    try{ transComponentStore(st) }
    catch {
      case e: NoSuchElementException =>
        println("Impossible state "+st+" of type "+familyNames(st.family)+
                  " encountered.\n"+
                  "This probably means that the state spaces are not symmetric.")
        sys.exit
    }
}

