package ViewAbstraction

/** A class to allow threads to concurrently put values into a buffer,
  * and then for a sequential get to be performed. */
trait ConcBuffer[A]{
  /** Put x into the buffer. */
  def put(x: A) 

  /** Sequentially get the contents of the buffer. */
  def get: Array[A] 
}


// =======================================================

import scala.collection.mutable.ArrayBuffer

class ShardedConcBuffer[A: scala.reflect.ClassTag](shards: Int = 128)
    extends Sharding(shards) with ConcBuffer[A]{
  /** The buffers holding the data. */
  private val buffers = Array.fill(shards)(new ArrayBuffer[A])


  /** Put x into the buffer. */
  def put(x: A) = {
    val h = hashOf(x); val sh = shardFor(h)
    buffers(sh).synchronized{ buffers(sh) += x }
  }

  /** Sequentially get the contents of the buffer. */
  def get: Array[A] = {
    val len = buffers.map(_.length).sum
    val result = new Array[A](len)
    var next = 0
    for(i <- 0 until shards){
      val len1 = buffers(i).length
      buffers(i).copyToArray(result, next, len1)
      buffers(i) = null // for GC
      next += len1
    }
    result
  }

}
