package ViewAbstraction

import scala.collection.mutable.ArrayBuffer

/** Object to produce debugging output.
  * @param system the system we are debugging.
  * @param k the abstraction parameter used.
  * @param l the length of concrete views.
  * @param aShapes the shapes of abstract views
  * @param sysAbsViews the abstract views encountered during the search.
  * @param initViews the initial views (of size k). */
class Debugger(
  system: System, threeWaySync: Boolean, aShapes: List[Array[Int]],
  sysAbsViews: SystemViewSet, initViews: List[SystemView], isDeadlock: Boolean){

  // Following are used only in assertions, I think
  val k = aShapes.head.sum
  val l = if(threeWaySync) k+2 else k+1

  // Parameters used in printing: indent is the width of the left-hand column;
  // pad is a blank left-hand column.
  private val indent = 5; private val pad0 = " "*(indent-3);
  private val pad = " "*indent
  /** Right pad st to the width of the left-hand column. */
  def rPad(st: String) = st.padTo(indent, ' ')
 
  /** Information about a trace found. 
    * For each i in [0..plies),
    * abss(i) =gamma=> befores(i) -events(i)-> afters(i) =alpha=>
    * abs1s(i) == abss(i+1)
    * and abss(plies) =gamma=> befores(plies), where plies = events.size.
    */
  private case class TraceInfo(
    abss: ArrayBuffer[SystemView], befores: ArrayBuffer[SystemView],
    events: ArrayBuffer[EventInt],
    afters: ArrayBuffer[SystemView], abs1s: ArrayBuffer[SystemView])
  {
    val plies = events.size
    assert(abss.size == plies+1 && befores.size == plies+1 &&
             afters.size == plies && abs1s.size == plies)

    /** Print this trace. */
    def print(toError: Boolean = false) = {
     println
      println(pad+abss(0)+"\n"+pad+"=gamma=>\n"+rPad("0: ")+befores(0))
      for(i <- 0 until plies){
        println(pad0+"-"+system.showEvent(events(i))+"->\n"+pad+afters(i))
        if(afters(i) != befores(i+1)){
          println(pad+"=alpha=>\n"+pad+abs1s(i))
          if(abss(i+1) != abs1s(i)) println(pad+"== "+abss(i+1))
          println(pad+"=gamma=>\n"+rPad((i+1).toString+": ")+befores(i+1))
        }
      }
      if(toError){ 
        if(isDeadlock) println(pad+"deadlocked") else println(pad+"=error=>")
      }
    }
  } // end of TraceInfo

  /* ============================================ */

  /** Produce debugging information for last.
    * @param plies the number of plies in the search, i.e. the number of events
    * that happened before the state last was reached. */
  def apply(last: SystemView): Unit = {
    println("debugging")
    apply1(last, null)
  }

  /** Produce debugging information for last.
    * @param penultimate if non-null, a SystemView of size k+1 that contributed
    * to last, and that should be included in the trace. */
  def apply1(last: SystemView, penultimate: SystemView): Unit = {
    // Find trace leading to error state.
    val tr = findTrace(last, penultimate)
    // Print trace
    tr.print(penultimate == null)
    // Interactively get next thing to do
    def help =
      println(
        "Type: a number from 0 to "+tr.plies+
          " to expand the sources of that state;\n"+
          "'u' to go up a level; 's' to list servers; 'q' to quit.")
    var done = false
    while(!done){
      val input = scala.io.StdIn.readLine("> ")
      input match{
        case "q" => println("Goodbye."); sys.exit
        case "u" => done = true // backtrack up the stack of traces
        case st if st.nonEmpty && st.forall(_.isDigit) =>
          expandGamma(tr.befores(st.toInt))
          tr.print(penultimate == null) // re-print trace on return
        case "s" => showServers
        case "h" | "?" | "help" => help
        case _ => println("Invalid command."); help
      }
    }
  }

  // Flag to make findTrace print information during the backwards search
  private val verbose = false

  /** Find the trace leading to state last. 
    * @param penultimate if non-null, a SystemView of size k that contributed
    * to last, and that should be included in the trace. */
  private def findTrace(last: SystemView, penultimate: SystemView = null)
      : TraceInfo = {
    var before = last
    assert(before.componentView.length == l)
    // assert(checkLength(before.componentView, 1))
    assert(penultimate == null || penultimate.componentView.length == k)
    // assert(penultimate == null || checkLength(penultimate.componentView, 0))
    var done = false
    // init records whether we should select penultimate as the source of the
    // next gamma transition.
    var init = penultimate != null
    // Build up trace in reverse.
    val befores, abss, abs1s, afters = new ArrayBuffer[SystemView]
    val events = new ArrayBuffer[EventInt]
    befores += last
    // Inv: for each i in [0..events.length),
    //        befores(i) -events(i)-> afters(i) =alpha=>
    //        abs1s(i) == abss(i+1) =gamma=> befores(i+1)
    while(!done){
       // before was created by gamma using the elements of absList
      val absList =
        if(init) { init = false; List(penultimate) }
        else
          Views.alpha(aShapes, before.componentView).map(absCpts =>
            SystemView.mkSystemView(before.servers, absCpts))
      // after =alpha=> abs1 == abs =gamma=> before, and this is the latest
      // contributor to before to be found.
      val (abs,abs1,after,ply) = absList.map(getStep).maxBy(_._4)
      // val abs =
      //   if(init) { init = false; penultimate }
      //   else{
      //     val absCpts = Views.alpha(k, aShapes, before.componentView).head
      //     system.mkSystemView(before.servers, absCpts)
      //   }
      abss += abs
      // val (abs1,after,ply) = getStep(abs)
      if(ply < 0) done = true
      // if(initViews.contains(abs1)) done = true
      else{
        abs1s += abs1
        // prev1 was created by alpha from prev2
        //val (after,_) = SystemView.alphaLog(abs1)
        //if(verbose) println("<-alpha=\n"+after)
        afters += after 

        // The checker found the transition before -e-> after
        val (b, e) = after.getPred; before = b
        // val (b, e) = system.debugLog(after); before = b
        // assert(after.getPred == (b,e))
        assert(before.componentView.length == l)
        // assert(checkLength(before.componentView, 1))
        if(verbose) println("<-"+system.showEvent(e)+"-\n"+before)
        befores += before; events += e 
      }
    }
    if(verbose) println
    TraceInfo(abss.reverse, befores.reverse, events.reverse,
              afters.reverse, abs1s.reverse)
  }

  /** Given an abstraction abs, return (abs, abs1, after, ply) such that on ply
    * ply, after =alpha=> abs1 equiv abs. */
  def getStep(abs: SystemView): (SystemView, SystemView, SystemView, Int) = {
    assert(abs.componentView.length == k)
    if(verbose) println("<=gamma=\n"+abs)
    // abs1 is equivalent to abs in sysAbsViews.
    val abs1 = sysAbsViews.getRepresentative(abs)
    assert(abs1 == abs) // but they will be different objects
    if(initViews.contains(abs1)) (abs, abs1, null, -1)
    else{
      // val (after,ply) = SystemView.alphaLog(abs1)
      // assert(abs1.getPred == (after,ply))
      val (after,ply) = abs1.getPred
      if(verbose) println("<-alpha=\n"+after)
      (abs, abs1, after, ply)
    }
  }


  /** Print the server names. */
  private def showServers = {
    val serverNames = system.serverNames
    for(i <- 0 until serverNames.length) println(rPad(i+":")+serverNames(i))
  }

  /** Show all SystemViews of size k that contributed to conc (of size l). */
  private def expandGamma(conc: SystemView) = {
    val cmpts = conc.componentView
    assert(cmpts.length == l)
    // assert(checkLength(cmpts, 1))
    println("Expanding "+conc)
    // The abstract views of the components
    val absViews = Views.alpha(aShapes, cmpts)
    // (for(i <- 0 until k+1) yield cmpts.take(i)++cmpts.drop(i+1)).toList
    // Create all abstract system views from these
    val abss: List[SystemView] =
      absViews.map(SystemView.mkSystemView(conc.servers, _))
    def printAbss = for(i <- 0 until abss.length) println(rPad(i+": ")+abss(i))
    printAbss
    // Interactively get next thing to do
    def help =
      println(
        "Type: a number from 0 to "+(abss.length-1)+
          " to obtain the trace leading to that state;\n"+
          "'u' to go up a level; 's' to list servers; 'q' to quit.")
    var done = false
    while(!done){
      val input = scala.io.StdIn.readLine("> ")
      input match{
        case "q" => println("Goodbye."); sys.exit
        case "u" => done = true // backtrack up the stack of traces
        case st if st.nonEmpty && st.forall(_.isDigit) =>
          apply1(conc, abss(st.toInt))
          printAbss // re-print choices on return
        case "s" => showServers
        case "h" | "?" | "help" => help
        case _ => println("Invalid command."); help
      }
    }
  }

  /* IMPROVE: The debugging needs to be improved.  Try to find consecutive
   * alpha/gamma steps to minimise the change, i.e. either giving the same
   * view or one that is equivalent, if possible. */

}
