package ViewAbstraction

/** Abstractly, classes that extend ExtenderMap represent mappings from 
  * SubSystemView to sets of States.  
  * 
  * There is an assumption that adds do not overlap with gets or other query
  * operations.*/
trait ExtenderMap{
  type SubSystemView = ExtenderMap.SubSystemView

  /** Add each state in sts, of family f, to the set of States associated with
    * ssv. */
  def add(f: Family, ssv: SubSystemView, sts: List[State])

  /** An iterator over the states of family f associated with ssv. */
  def get(f: Family, ssv: SubSystemView): Iterator[State]

  /** The size of the domain of the mapping. */
  def size: Int

  /** The number of States in the range of the mapping. */
  def stateCount: Int
}


// -------------------------------------------------------

object ExtenderMap{
  type View = Views.View

  /** A subview of a system view; used in the domain of the store mapping. */
  class SubSystemView(val servers: ServerStates, val view: View){
    override def equals(that: Any) = {
      val ssv = that.asInstanceOf[SubSystemView]
      ssv.servers == servers && ssv.view.sameElements(view)
    }

    override def hashCode = {
      var h = servers.hashCode
      for(st <- view) h = h*73 + st.hashCode
      h
    }

    override def toString =
      "SubSystemView("+servers.servers+", "+view.mkString("<", ",", ">")+")"
  } // end of SubSystemView
}

// -------------------------------------------------------

import java.util.concurrent.atomic.AtomicInteger
import scala.collection.mutable.Set

/** A simple implementation of the ExtenderMap trait. */
class ExtenderMap1 extends ExtenderMap{
  /** The store holding the mapping so far.  Each entry ssv -> states in
    * store(f) means that for each st in states, st is associates with ssv;
    * further, st is of family f. */
  private val store =
    Array.fill(numFamilies)(
      new MyLockFreeReadHashMap[SubSystemView, Set[State]]())

  /** The number of States in the range of store.  */
  private val theStateCount = new AtomicInteger(0)

  /** Add each state in sts, of family f, to the set of States associated with
    * ssv. */
  def add(f: Family, ssv: SubSystemView, sts: List[State]) = {
    val set = store(f).getOrElseUpdate(ssv, Set[State]())
    for(st <- sts)
      if(set.synchronized{set.add(st)}) theStateCount.getAndIncrement
  }

  def get(f: Family, ssv: SubSystemView): Iterator[State] = 
    store(f)(ssv).iterator

  def size = store.map(_.size).sum

  def stateCount = theStateCount.get
}

// -------------------------------------------------------

/** An implementation of ExtenderMap based on sharding. 
  * @param shards the number of shards
  * @param initLength the initial size of each shard. */
class ShardedExtenderMap(
  shards : Int = LockFreeReadHashSet.powerOf2AboveNumThreads*16,
  initLength: Int = 32) 
extends  Sharding(shards) with ExtenderMap{

  /** Each Shard represents part of the mapping for one family. */
  private class Shard{
    /** Current number of slots. */
    var size = initLength

    /** The hashes of the mapping.  Values of 0 indicate an empty slot. */
    var hashes = new Array[Int](size)

    /** The keys of the mapping. */
    var keys = new Array[SubSystemView](size)

    /** The current number of entries, i.e. the domain of the size of the map. */
    var count = 0

    /** Maximum load factor before resizing. */
    private val MaxKeyLoad = 0.666

    /** Threshold for next resizing, based on size. */
    @inline private def mkThreshold = (size*MaxKeyLoad).toInt

    /** Threshold at which global resizing should be performed. */
    private var threshold = mkThreshold

    /** The corresponding states. */
    private var states = new Array[Array[State]](size)

    /** The number of entries in each entry of states. */
    private var rangeSize = new Array[Int](size)

    /** The threshold for resizing each entry in states. */
    private var rangeThreshold = new Array[Int](size)

    /** Total number of states. */
    var stateCount = 0

    /** Maximum load factor of each range array before resizing. */
    private val MaxRangeLoad = 0.7

    /** Threshold for resizing range array of size n. */
    @inline private def mkRangeThreshold(n: Int) = (n*MaxRangeLoad).toInt

    /** Bit map for selecting slot. */
    private var mask = size-1

    /** Find possible slot for ssv with hash h. */
    @inline private def find(ssv: SubSystemView, h: Int): Int = {
      var i = h & mask // ; val start = i
      while({ val h1 = hashes(i); h1 != 0 && (h1 != h || keys(i) != ssv) }){
        i = (i+1) & mask // (i+1)%size
      }
      i
    }

    /** Resize this. */
    @inline private def resize() = {
      val oldSize = size; size += size; threshold = mkThreshold; mask = size-1
      val oldHashes = hashes; hashes = new Array[Int](size)
      val oldKeys = keys; keys = new Array[SubSystemView](size)
      val oldStates = states; states = new Array[Array[State]](size)
      val oldRangeSize = rangeSize; rangeSize = new Array[Int](size)
      val oldRangeThreshold = rangeThreshold
      rangeThreshold = new Array[Int](size)
      // Copy data
      for(i <- 0 until oldSize; if oldHashes(i) != 0){
        val ssv = oldKeys(i); val h = oldHashes(i); val j = find(ssv, h)
        // assert(hashes(j) == 0) 
        hashes(j) = h; keys(j) = ssv; states(j) = oldStates(i)
        rangeSize(j) = oldRangeSize(i); rangeThreshold(j) = oldRangeThreshold(i)
      }
    }

    /** Try to add sts to the set associated with ssv.  Return true if
      * successful (false implies the thread must retry). */
    def add(ssv: SubSystemView, h: Int, sts: List[State]): Unit = synchronized{
      // Find position for ssv
      var i = find(ssv, h)
      // Add ssv to the mapping if necessary
      if(hashes(i) == 0){
        if(count >= threshold){ resize(); i = find(ssv, h) }
        keys(i) = ssv; states(i) = initRangeArray(i, sts.length)
        rangeThreshold(i) = mkRangeThreshold(states(i).length)
        hashes(i) = h; count += 1
      }
      // assert(keys(i) == ssv) 
      // Add each element of sts to states(i)
      for(st <- sts) addState(i, states(i), st)
    } // end of add

    /** Create an initial array suitable for holding n states. */
    @inline private def initRangeArray(i: Int, n: Int): Array[State] = { 
      val m = (n/MaxRangeLoad).toInt // m >= n
      var k = 1; while(k <= m) k += k
      // k > n and k is a power of 2
      new Array[State](k)
    }

    /** Add state st to myStates = states(i), resizing if necessary. */
    @inline private def addState(i: Int, myStates: Array[State], st: State)
        : Unit = {
      // Resize if necessary
      if(rangeSize(i) >= rangeThreshold(i)){
        val newStates = resizeRange(myStates); states(i) = newStates
        rangeThreshold(i) = mkRangeThreshold(newStates.length)
        addState(i, newStates, st)
      }
      else{
        // Search for slot for st
        val h1 = hashOf(st); val mask = myStates.length-1; var j = h1 & mask
        while(myStates(j) != null && myStates(j) != st) j = (j+1) & mask
        if(myStates(j) == null){  // add st
          myStates(j) = st; rangeSize(i) += 1; stateCount += 1
        }
      }
    }

    /** Resize the range array myStates. */
    @inline private def resizeRange(myStates: Array[State]): Array[State] = {
      val oldSize = myStates.length; val newSize = 2*oldSize
      val newStates = new Array[State](newSize); val mask = newSize-1
      // Copy states across
      for(i <- 0 until oldSize){
        val st = myStates(i)
        if(st != null){
          val h = hashOf(st); var j = h & mask
          while(newStates(j) != null) j = (j+1) & mask
          newStates(j) = st
        }
      }
      newStates
    }

    def get(ssv: SubSystemView, h: Int): Iterator[State] = {
      val i = find(ssv, h); // assert(keys(i) == ssv) 
      new Iterator[State]{
        private val myStates = states(i); val n = myStates.length
        // index of next state, or n if none left
        private var j = 0; advance()

        // Advance to next state. 
        @inline private def advance() = 
          while(j < n && myStates(j) == null) j += 1

        def hasNext = j < n

        def next = { val st = myStates(j); j += 1; advance(); st }
      }
    }
  } // end of TableState 

  // ----------

  private val theShards = Array.fill(numFamilies, shards)(new Shard)

  def add(f: Family, ssv: SubSystemView, sts: List[State]) = {
    val h = hashOf(ssv); val shard = theShards(f)(shardFor(h))
    shard.add(ssv, h, sts)
  }

  def get(f: Int, ssv: SubSystemView): Iterator[State] = {
    val h = hashOf(ssv); val shard = theShards(f)(shardFor(h))
    shard.get(ssv, h)
  }

  def size: Int = {
    var c = 0
    for(f <- 0 until numFamilies; sh <- theShards(f)) c += sh.count
    c
  }

  def stateCount: Int =  {
    var c = 0
    for(f <- 0 until numFamilies; sh <- theShards(f)) c += sh.stateCount
    c
  }

}
