package ViewAbstraction

import uk.ac.ox.cs.fdr._
import scala.collection.JavaConverters._
import scala.collection.mutable.Map

/** An object that creates an FDR session, and offers various operations on
  * it. 
  * @param fname the name of the file. */
class FDRSession(fname: String){
  /** The FDR Session object. */
  private val session = new Session

  session.loadFile(fname)

  /** The Cancellor, used in queries on the session. */
  private val canceller = new Canceller

  // Evaluation of non-process values. 

  /** Use the FDR session to evaluate the expression st. */ 
  def eval(st: String): String =
    session.evaluateExpression(st, canceller).result

  /** Convert internal representation of event into the corresponding 
    * String. */
  def uncompileEvent(event: Int): String =
    session.uncompileEvent(event).toString()

  /** Convert String representing event into internal representation. */
  private def compileEvent(st: String): Int =
    session.compileEvent(st, canceller).result.toInt

  /** Convert a String, representing a set, into a list of Strings. 
    * 
    * Note: avoid calling this on type names; use getTypeValues, instead. */
  def setStringToList(st: String): List[String] = 
    session.evaluateIterableExpression(st, canceller).result.asScala.toList

  /** Int representing the value st from the symmetry sub-type. */
  def symmetryValue(st: String): Int =
    session.symmetryValue(st, canceller).result.toInt

  /** Number of visible events. */
  def numEvents = eval("card(Events)").toInt

  // StringList to use in evalProc
  private val types = new StringList

  /** The values in a type; memoized in getTypeValues. */
  private val typeValues = Map[String, List[String]]()

  /** Get the values in a type. */
  def getTypeValues(typeName: String): List[String] = 
    typeValues.getOrElseUpdate(typeName, setStringToList(typeName))

  /** Get and record the type that includes the subtype typeName. */
  def getSuperType(typeName: String) = {
    val typeValues = getTypeValues(typeName) // setStringToList(typeName)
    val superType = session.typeOfExpression(typeValues.head).result
    types.add(superType)
    superType
  }

  /** Get the name of the type represented by t. */
  def typeName(t: Long): String =
    session.symmetryTypeName(java.math.BigInteger.valueOf(t))


  // ======== Events and processes

  /** Get the int representing type typeName internally.
    * @param procName the name of a process parameterised by typeName, so 
    * procName(x) is a valid process for each x in typeName. */
  def getTypeInt(typeName: String): Long = {
    val tVals = getTypeValues(typeName)
    println(typeName+" -> "+tVals)
    def proc(x: String) =
      "let P_(x_) = error -> if x_ == "+x+" then SKIP else STOP within P_("+x+")"
    val machine = evalProc(proc(tVals.head))
    val variableVals = machine.variableValues(machine.rootNode).asScala.toList
    assert(variableVals.length == 1)
    variableVals.head.getType.longValue
    // This is a little inefficient, as we re-build this machine later.
    // val machine = evalProc(procName+"("+tVals.head+")")
    // val root = machine.rootNode
    // val variableVals = machine.variableValues(root).asScala.toList
    // // assert(variableVals.length == 1, variableVals.map(_.getValue.intValue))
    // assert(vv1.head.getType.longValue == variableVals.head.getType.longValue)
    // variableVals.head.getType.longValue
  }

  /** Use FDR to evaluate process expression st.
    * Pre: an earlier call to getSuperType initialised types. */
  def evalProc(st: String): Machine =
    session.evaluateProcessForSymmetry(
      st, SemanticModel.Traces, types,  canceller).result

  /** Map giving the String for each event number. */
  private val eventStringMap = Map[Int, String]()

  /** Given internal representation of event, get corresponding String. */
  def eventToString(n: Int): String = eventStringMap(n)

  /** Map giving the event number for a String representing an event. */
  private val stringEventMap = Map[String, Int]()

  /** Get internal representation of event represented as String. */
  def eventToInt(st: String): Int = stringEventMap.get(st) match{
    case Some(n) => n
    case None =>
      val n = compileEvent(st); eventStringMap += n -> st
      // println("eventToInt "+n+" "+st)
      stringEventMap += st -> n; n
  }

  /** Add an event to eventStringMap and stringEventMap. */
  def logEvent(event: Int) = {
    if(!eventStringMap.contains(event)){
      val st = session.uncompileEvent(event).toString()
      //println(st)
      eventStringMap += event -> st; stringEventMap += st -> event
    }
  }

  /** Initialise eventStringMap and stringEventMap.
    * @param eventsSize the number of events (including tau and tick) plus one.
    */
  def initEvents(eventsSize: Int) = {
    evalProc("RUN(Events)") // necessary or else uncompileEvent barfs.
    for(e <- 1 until eventsSize) logEvent(e)
  } 

}
