package ViewAbstraction

import uk.ac.ox.cs.fdr._
import scala.collection.JavaConverters._
import scala.collection.mutable.{Map,Stack,Set,ArrayBuffer}

/** Class that builds transition systems, based on FDR.
  * @param fdrSession the FDR session object
  * @param superTypeNames the names of distinguished supertypes.  */
class FDRTransitionMap(
  fdrSession: FDRSession, superTypeNames: Array[String])
{

  // ========= Information about types

  // assert(familyTypeNames.distinct.length == numTypes &&  
  //          superTypeNames.length == familyTypeNames.length)
  
  /** A map for a particular supertype, mapping, for each element x of that
    * supertype, the representation of x inside FDR to the representation used
    * here; elements of the subtype map onto an initial segment of the
    * naturals, and other (distinguished) values map onto negative ints. */
  private type IdRemap = Map[Int, Int]

  /** A map giving an IdRemap for each type, indexed by the representation of
    * the types inside FDR. */
  private val idRemaps = Map[Long, IdRemap]() 

  /** Array giving the number of elements of each indexing subtype. */
  // val idSizes = new Array[Int](numTypes)

  /** A map for a particular type, mapping from the representation used here to
    * the name from the script. */
  private type NameMap = Map[Identity, String]

  /** A map giving a NameMap for each type. */
  private val theNameMap = Map[Type, NameMap]()

  def getNameMap = theNameMap

  /** Map from the Longs used to represent types in FDR to the Ints used here.  
    * The latter are in the range [0..numTypes). 
    * 
    * The entry for type t is built by typeMap, the first time a parameter of
    * this type is encountered. */
  private val theTypeMap = Map[Long, Int]()

  /** Given the Long used to represent a type within FDR, return the
    * corresponding Int used here.  Also, if this is the first time we've seen
    * this type, calculate and store information about the type, updating
    * theTypeMap, fdrTypeIds, idRemaps and typeSizes.  */
  private def typeMap(t: Long): Type = theTypeMap.get(t) match{
    case Some(i) => i
    case None =>
      // This is the first time we've encountered this type
      val superTypeName = fdrSession.typeName(t)
      val i = superTypeNames.indexOf(superTypeName); assert(i >= 0)
      println(superTypeName+"\t"+t+"\t"+i)
      theTypeMap += t -> i 
      val (idRemap, nameMap, typeSize) =
        buildIdRemap(familyTypeNames(i), superTypeName)
      idRemaps += t -> idRemap; theNameMap += i -> nameMap;
      typeSizes(i) = typeSize; superTypeSizes(i) = idRemap.size
      distinguishedSizes(i) = superTypeSizes(i) - typeSize
      println("Supertype size = "+idRemap.size)
      println("Distinguished values = "+distinguishedSizes(i))
      i
  }

  def fdrTypeToType(t: Long): Type = typeMap(t)

  /** Build information about type typeName, with supertype superType. 
    * @return a triple: (1) an IdRemap, mapping the representation of each 
    * value inside FDR to the value used here; (2) a map from the values used
    * here to the names in the script; (3) the number of elements of the type
    * (excluding distinguished values of the supertype).*/
  private def buildIdRemap(typeName: String, superType: String)
      : (IdRemap, NameMap, Int) = {
    // println("buildTypeMap: typeName = "+typeName+"; superType = "+superType)
    val idRemap = Map[Int, Int]() // the result
    val nameMap = Map[Int, String]() // map from values used here to script names
    // Values in the type and supertype
    val superTypeVals = fdrSession.getTypeValues(superType)
    val typeValues = fdrSession.getTypeValues(typeName)
    // val superTypeVals = fdrSession.setStringToList(superType)
    // val typeValues = fdrSession.setStringToList(typeName)
    // Next ints to use for ids and distinguished values
    var nextId = 0; var nextDistinguished = -1

    // Build the mapping for each value of superType in turn.
    for(st <- superTypeVals){
      val id = fdrSession.symmetryValue(st) // Int representing st inside FDR
      if(typeValues.contains(st)){ idRemap += id -> nextId; nextId += 1 }
      else{ idRemap += id -> nextDistinguished; nextDistinguished -= 1 }
      nameMap += idRemap(id) -> st
      println(id+": "+st+" -> "+idRemap(id))
    }

    (idRemap, nameMap, nextId)
  }

  // ========= Build the transition systems.

  // IMPROVE: we create the transition function in one form, then transform it
  // to a different form.  Can this be improved?

  /** A map representing a transition system.  Given a destination src, the map
    * returns a list of events e and states dst such that src -e-> dst.  This
    * list is sorted by events. */ 
  type TransMap = Map[State, List[(Int, State)]]

  /** A map giving the types of the parameters of each state. */
  private val stateTypeMap0 = Map[ControlState, Array[Type]]()

  /** Maximum control state used. */
  var maxCS = -1

  /** Convert a Node into a State, using the node's state group as the
    * ControlState, and the node's variable values as the Identities,
    * prepended with id if non-negative and the identity is not already there,
    * and remapped using idRemap.
    * @param family the family of components from which this is taken, or -1 
    * for a server.
    * @param id the representation inside FDR of this process's identity, 
    * or -1 if this isn't a replicated component.
    * @param typeId the type identifier for the type of id. */
  private def nodeToState
    (machine: Machine, node: Node, family: Int, id: Int, typeId: Long)
      : State = {
    val variableVals = machine.variableValues(node).asScala.toList
    // Types and values for each variable map.
    val args0: List[(Long, Int)] =
      variableVals.map(vv => (vv.getType.longValue, vv.getValue.intValue))
    // Add in id if relevant
    val args1 =
      if(id >= 0){
        if(args0.isEmpty || args0.head != (typeId, id)){
          println("adding identity"); (typeId, id)::args0 
        }
        else args0
      }
      else args0
    val cs = machine.stateGroup(node).toInt
    maxCS = maxCS max cs
    // assert(cs >= 0, cs+"\t"+machine.stateGroup(node))
    // val args2 = args1.map{ case (t,x) => (typeMap(t), idRemaps(t)(x)) }
    val ids: Array[Identity] = args1.map{ case (t,x) => idRemaps(t)(x) }.toArray
    stateTypeMap0.get(cs) match{
      case Some(ts) => { } // assert(types.sameElements(ts))
      case None => 
        val types: Array[Type] = args1.map{ case (t,x) => typeMap(t) }.toArray
        stateTypeMap0 += cs -> types
    }
    State(family, cs, ids, true)
  }

  // Flag to cause transitions to be printed
  private val verbose = false

  /** Explore machine, adding all transitions to transMap.  seen contains those
    * states already explored.  Also log all events with fdrSession. 
    * @param family the family of components from which this is taken, or -1 
    * for a server.
    * @param id the identity of the process represented by this machine if it
    * is a component, or -1 otherwise.
    * @param typeId the FDR type identifier for the type of id 
    * (if non-negative). 
    * @return the root state as a State. */
  def augmentTransMap(
    transMap: TransMap, seen: Set[State], machine: Machine,
    family: Int, id: Int, typeId: Long, verbose: Boolean = false)
      : State = {
    val root = machine.rootNode
    val rootState = nodeToState(machine, root, family, id, typeId)
    if(!seen.contains(rootState)){
      // We'll perform a depth-first search.  stack will store Nodes, and the
      // corresponding State values, that have to be expanded.
      val stack = new Stack[(Node, State)]
      stack.push((root, rootState)); seen += rootState

      while(stack.nonEmpty){
        val (node, st) = stack.pop
        // Build up transitions from node in trans
        var trans = ArrayBuffer[(Int, State)]()
        var lastE = -1
        for(t <- machine.transitions(node).iterator.asScala) {
          val e = t.event.toInt; fdrSession.logEvent(e)
          assert(lastE <= e); lastE = e // check events non-decreasing
          val dst = t.destination
          val dstState = nodeToState(machine, dst, family, id, typeId)
          trans += ((e, dstState))
          if(verbose)
            println(st.toString00+" -"+fdrSession.eventToString(e)+"-> "+
                      dstState.toString00)
          // assert(dstState._1 >= 0)
          if(!seen.contains(dstState)){
            stack.push((dst, dstState)); seen += dstState
          }
        } // end of for loop
        transMap += st -> trans.toList
      }
    }
    else println(rootState+" already seen")
    rootState
  }

  /** Build a transition map corresponding to name.
    * @param alpha the alphabet of this process in the system; transitions 
    * outside this alphabet should be dropped.
    * @param family the family of components from which this is taken, or -1 
    * for a server.
    * @return a triple consisting of (1) the initial state, as a State, (2) the 
    * transition map as a map from States to a pair of parallel arrays, giving
    * each event and list of post-states. */
  def buildTransMap(name: String, alpha: Set[EventInt], family: Int)
      : (State, Map[State, (Array[EventInt], Array[List[State]])]) = {
    // println("buildTransMap: "+name)
    val machine: Machine = fdrSession.evalProc(name)
    val transMap = Map[State, List[(Int, State)]]()
    val seen = Set[State]()
    val init = augmentTransMap(transMap, seen, machine, family, -1, -1)
    
    // Now convert into more convenient form, giving a map from states to a
    // pair of parallel arrays, giving each event and list of post-states.
    val newTransMap = Map[State, (Array[EventInt], Array[List[State]])]()
    for(s <- seen){
      var transList: List[(EventInt, State)] = transMap(s)
      val es = new ArrayBuffer[EventInt]
      val nexts = new ArrayBuffer[List[State]]
      while(transList.nonEmpty){
        val e = transList.head._1
        val (matches, rest) = transList.span(_._1 == e)
        transList = rest
        if(matches.nonEmpty){
          if(alpha.contains(e) || e == Tau){ 
            // println(e+" -> "+matches.map(_._2.toString00))
            es += e; nexts += matches.map(_._2) 
          }
          // else println("transition on "+fdrSession.eventToString(e)+" dropped")
        }
      }
      es += Int.MaxValue // sentinel
      newTransMap(s) = (es.toArray, nexts.toArray)
    }
 
    (init, newTransMap)
  }

  /** Create stateTypeMap, giving the types of parameters of each state.  This
    * should be called after the transition maps of the components and servers
    * have been created. */
  def createStateTypeMap = {
    val maxCS = stateTypeMap0.keysIterator.max
    // following two variables in package
    val minCS = stateTypeMap0.keysIterator.min
    println("minCS = "+minCS+"; maxCS = "+maxCS)
    val stateTypeMapArray = new Array[Array[Type]](maxCS-minCS+1)
    for((cs,ts) <- stateTypeMap0.iterator) stateTypeMapArray(cs-minCS) = ts
    State.setStateTypeMap(stateTypeMapArray, minCS)
  }


}
