package ViewAbstraction

// import ox.cads.util.Profiler

/** Trait for a map storing States, keyed by Family, ControlStates and
  * Identitys.  */
trait MyStateMap{
  /** The number of elements in this. */
  def size: Int

  /** Get the State corresponding to (family, cs, ids), or create one if there
    * is no such. 
    * @param isNew is this being done during compilation?  If so, the ids 
    * are bound to be values from the script. */
  @inline def getOrAdd(
    family: Int, cs: ControlState, ids: Array[Identity], isNew: Boolean)
      : State

  /** Add st to the store, or get the equivalent one if it exists. */
  @inline def add(st: State): State
}

// =======================================================

/** A hash map storing States, keyed by ControlStates and Identitys. 
  *  Not thread safe. */
class MyStateHashMap(initLength: Int = 4096) extends MyStateMap{

  // check initSize is a power of 2
  checkPow2(initLength)

  /** The current number of spaces in the table. */
  private var n = initLength

  /** Bit mask to produce a value in [0..n). */
  private var mask = n-1

  /** The current number of entries. */
  private var count = 0

  /** The number of elements in this. */
  def size = count

  /** Maximum load factor before resizing. */
  private val MaxLoad = 0.5

  /** Threshold at which resizing should be performed. */
  private var threshold = (initLength*MaxLoad).toInt

  /** The hashes of the mapping. */
  private var hashes = new Array[Int](initLength)

  /** The control states. */
  private var controlStates = new Array[ControlState](initLength)

  /** The identites. */
  private var identities = new Array[Array[Identity]](initLength)

  /** The states. */
  private var states = new Array[State](initLength)

  /** Hash of x; guaranteed not to be 0. */
  @inline private def hashOf(cs: ControlState, ids: Array[Identity]) = {
    var h = cs; var i = 0; var n = ids.length
    while(i < n){ h = h*41+ids(i); i += 1 }
    h ^= ((h >>> 20) ^ (h >>> 12))
    h ^= (h >>> 7) ^ (h >>> 4)
    if(h == 0) 12344577 else h
  }

  /* This represents the mapping 
   * { (controlStates(i), identities(i)) -> states(i) | hashes(i) != 0 }.
   * 
   * Each element is placed in the first empty (null) position
   * starting from its hash.  Its hash is placed in the corresponding
   * position in hashes. */

  /** Resize the hash table. */
  @inline private def resize() = {
    val oldN = n; n += n; threshold += threshold; mask = n-1
    // println("resize to "+n)
    val oldHashes = hashes; hashes = new Array[Int](n)
    val oldControlStates = controlStates
    controlStates = new Array[ControlState](n)
    val oldIdentities = identities; identities = new Array[Array[Identity]](n)
    val oldStates = states; states = new Array[State](n)
    var j = 0 // index into old arrays
    while(j < oldN){
      val h = oldHashes(j)
      if(h != 0){ // add old*(j) to new table
        var i = h & mask
        while(hashes(i) != 0) i = (i+1) & mask // (i+1)%n
        hashes(i) = h; controlStates(i) = oldControlStates(j)
        identities(i) = oldIdentities(j); states(i) = oldStates(j)
      }
      j += 1
    }
  }

  /** Find the position at which element (cs,ids) with hash h is stored
    * or should be placed. */
  @inline private def find(cs: ControlState, ids: Array[Identity], h: Int)
      : Int = {
    var i = h & mask
    while(hashes(i) != 0 &&
            (hashes(i) != h || controlStates(i) != cs ||
               ! identities(i).sameElements(ids)))
      i = (i+1) & mask // (i+1)%n
    i
  }

  /** Get the State corresponding to (cs, ids), or create one if there
    * is no such. */
  @inline def getOrAdd(
    family: Int, cs: ControlState, ids: Array[Identity], isNew: Boolean)
      : State = {
    // Profiler.count("getOrAdd")
    if(count >= threshold) resize()
    val h = hashOf(cs, ids); val i = find(cs, ids, h); val oldSt = states(i)
    if(oldSt == null){
      // assert(hashes(i) == 0 && states(i) == null)
      hashes(i) = h; controlStates(i) = cs; identities(i) = ids
      val newState = new State(family, cs, ids, isNew); states(i) = newState
      count += 1; newState
    }
    else{
      // assert(oldSt.cs == cs && oldSt.ids.sameElements(ids));
      oldSt
    }
  }

  /** An iterator over the states. */
  def iterator = new Iterator[State]{
    /** Index of the next element. */
    private var i = 0

    advance

    /** Advance i to next element. */
    private def advance = while(i < n && hashes(i) == 0) i += 1

    /** Does the iterator have a next element? */
    def hasNext = i < n

    /** The next element of the iterator. */
    def next = { val result = states(i); i += 1; advance; result }    
  }

  def add(st: State) = ???
}

// =======================================================

/** An implementation of a MyStateMap based upon a trie. 
  * @param numCS the number of control states. */
class MyTrieStateMap(numCS: Int, minCS: Int) extends MyStateMap{
  /** A Trie for each control state, offset by minCS. */
  private var tries = new Array[Trie](numCS)

  /* Each access to tries is synchronized on tries.  Each access to the
   * array from a MidTries or StateArrayTries is synchronized on that
   * array.  
   * 
   * Each entry is written at most once, changing it from null to
   * non-null.  Therefore a read of a non-null value does not require
   * obtaining the lock.  It's only necessary to obtain the lock if it
   * appears necessary to write a value.  However, in such cases it's
   * necessary to re-read the entry, to ensure another thread hasn't
   * set it to a non-null value.  This tactic speeds up the getOrAdd
   * operation by a factor of about 50! */ 

  /** The index in tries to use for control state cs. */
  @inline private def indexForCS(cs: ControlState) = cs-minCS

  private val SplitFreshVal = State.SplitFreshVal

  /* Each element of tries might be a StateTrie, corresponding to a
   * state with no parameters.  Otherwise, it is a tree of Tries, with
   * StateArrayTries at the bottom (at depth the arity of the control
   * state), and MidTries at every other depth.  
   * 
   * At the top level, the trie is indexed by control states, offset
   * by minCS, to allow for negative values.
   * 
   * At lower levels, each trie is index by the values of *two* identities (or
   * maybe one identity for the last ply), offset to allow for distinguished
   * values, and with fresh values (> SplitFreshVal) remapped to form a
   * contiguous range.  See function indexFor in getOrAdd.
   */

  /** The number of states created. */
  private val stateCount = new java.util.concurrent.atomic.AtomicInteger(0)

  /** The maximum fresh value allowed in any state.  FIXME. */
  private val MaxFresh = 3

  /** sizeFor(indexForCS(cs))(index) gives the number of values allowed for the
    * parameter at position index in control state cs.  It equals
    * superTypeSizes(State.stateTypeMap(cs)(index)) + MaxFresh. */
  private val sizeFor0 = Array.tabulate(numCS){ csi => 
    val types = State.stateTypeMap(csi+minCS)
    if(types != null) types.map(superTypeSizes(_)+MaxFresh)
    else null
    // the +MaxFresh above is to allow for identities mapped to values greater
    // than State.SplitFreshVal
  }

  /** Get the state corresponding to cs and ids, creating it if necessary. */
  @inline def getOrAdd(
    family: Int, cs: ControlState, ids: Array[Identity], isNew: Boolean)
      : State = {
    getOrMaybeAdd(family, cs, ids, null, isNew)
    // assert(newState.cs == cs)
    // assert(newState.ids.sameElements(ids), 
    //        newState.ids.mkString(", ")+"; "+ids.mkString(", ")) 
  }

  /** offSets(cs-minCS)(index) holds the number of distinguished values in the
    * type of the parameter index of cs. */
  private val offSets = 
    Array.tabulate(numCS){ cs => 
      val types = State.stateTypeMap(cs+minCS)
      if(types != null) types.map(distinguishedSizes)
      else null 
    }

  /** offSets(cs-minCS)(index) holds, for control state cs, for each index, the
    * size of the supertype - SplitFreshVal. */
  private val freshOffsets = 
    Array.tabulate(numCS){ cs =>
      val types = State.stateTypeMap(cs+minCS)
      if(types != null) types.map(superTypeSizes(_)-SplitFreshVal)
      else null 
    }

  /** Get the state corresponding to cs and ids; if none is stored, set
    * it to addState if non-null, or create a new state. */
  @inline private def getOrMaybeAdd(
    family: Int, cs: ControlState, ids: Array[Identity],
    addState: State, isNew: Boolean)
      : State = { 
    // Types for each element of ids.
    // val stateTypeMap = State.stateTypeMap(cs); 
    val len = ids.length
    val csIndex = indexForCS(cs)
    // offset into tries for normal values
    val theOffsets = offSets(csIndex)
    // offset into tries for fresh parameters
    val theFreshOffsets = freshOffsets(csIndex)
    // sizes of ranges for parameters
    val theSizes = sizeFor0(csIndex)

    // The contribution to the index into a trie corresponding to the
    // parameter at position index
    @inline def indexFor0(index: Int) = {
      val id = ids(index)
      // Second term below is the number of distinguished values of
      // the type of this identity, so gives an offset in each trie
      if(id < SplitFreshVal) id + theOffsets(index)
      else{
        assert(id < SplitFreshVal+MaxFresh) // FIXME
        id+theFreshOffsets(index)
      }
    }

    // Index into a trie corresponding to the parameters at positions idsIndex
    // and indIndex+1
    @inline def indexFor(idsIndex: Int) = 
      if(idsIndex == len-1) indexFor0(idsIndex)
      else indexFor0(idsIndex) * theSizes(idsIndex+1) + indexFor0(idsIndex+1)

    // Size for the trie corresponding to the parameters at positions idsIndex
    // and indIndex+1
    @inline def sizeFor(cs: ControlState, idsIndex: Int) = {
      // assert(idsIndex < len, s"idsIndex = $idsIndex; len = $len")
      if(idsIndex == len-1) theSizes(idsIndex)
      else theSizes(idsIndex) * theSizes(idsIndex+1)
    }

    var trie: Trie = null

    if(len == 0){
      trie = tries(csIndex)
      if(trie == null){
        val newState = 
          if(addState == null) new State(family, cs, ids, isNew) else addState
        tries.synchronized{
          if(tries(csIndex) == null){
            tries(csIndex) = StateTrie(newState)
            stateCount.getAndIncrement; newState
          }
          else{
            println("**Newly created**");
            tries(csIndex).asInstanceOf[StateTrie].st
          }
        } // end of synchronized block
      } // end of if(trie == null)
      else trie.asInstanceOf[StateTrie].st
        // assert(st.cs == cs && st.ids.sameElements(ids))
    } // end of if(len == 0)
    else{
      trie = tries(csIndex)
      if(trie == null){ // Create trie for this ControlState
        val newSize = sizeFor(cs, 0)
        trie = if(len <= 2) new StateArrayTrie(newSize) else new MidTrie(newSize)
        tries.synchronized{
          if(tries(csIndex) == null) tries(csIndex) = trie
          else trie = tries(csIndex) // recently written
        } // end of synchronized
      } // end of if trie == null

      // Iterate down to bottom level, creating new Tries as necessary
      var idsIndex = 0 // index into ids
      while(idsIndex < len-2){
        val nextTries = trie.asInstanceOf[MidTrie].tries
        val index = indexFor(idsIndex); trie = nextTries(index)
        if(trie == null){
          val newSize = sizeFor(cs, idsIndex+2)
          trie =
            if(idsIndex >= len-4) new StateArrayTrie(newSize)
            else new MidTrie(newSize)
          nextTries.synchronized{
            if(nextTries(index) == null) nextTries(index) = trie
            else trie = nextTries(index) // recently written
          } // end of synchronized
        }
        idsIndex += 2
      }

      // Get or create and store State in bottom level Trie
      val states = trie.asInstanceOf[StateArrayTrie].states
      val index = indexFor(idsIndex)
      // assert(index < states.length, 
      //        s"index = $index; states.length = ${states.length}; "+
      //          s"addState = $addState; ids = ${ids.mkString(",")}; "+
      //          (ids(0)-SplitFreshVal))
      val st = states(index)
      if(st != null){ 
        /* assert(st.family == family); */ 
        State.returnIdentityArray(ids)
        st
      }
      //assert(st.cs == cs && st.ids.sameElements(ids));
      else{
        val newState = 
          if(addState == null) new State(family, cs, ids, isNew) 
          else addState
        states.synchronized{
          if(states(index) != null) states(index) // recently written
          else{ states(index) = newState; stateCount.getAndIncrement; newState }
        } // end of synchronized
      }
    } // end of outer else
  }

  /** Add st to this. */
  def add(st: State) = getOrMaybeAdd(st.family, st.cs, st.ids, st, false)

  def size = stateCount.get
}


// =======================================================

/** A Trie, a node in the data structure used in MyTrieStateMap. */
trait Trie

/** A leaf trie, holding a state; only used at the top level for
  * states with no parameters. */
case class StateTrie(st: State) extends Trie

/** A node of a trie, holding other tries as children. */
class MidTrie(size: Int) extends Trie{
  val tries = new Array[Trie](size)
}

/** A leaf of a trie, holding size states as children. */
class StateArrayTrie(size: Int) extends Trie{
  val states = new Array[State](size)
}

// =======================================================
