package ViewAbstraction

import scala.collection.mutable.{Map, Set, ArrayBuffer}
// import ox.cads.util.Profiler
import java.util.concurrent.atomic.{AtomicInteger,AtomicLong}

/** Utility object to create concretizations of views.
  * @param concViews the set to which new views are added. 
  * @param aShapes the shapes of abstractions that we are extending. 
  * @param cShapes the shapes of concretizations that we're trying to create; 
  * each array sh represents sh(t) components of type t.
  */
class NewViewExtender(
  concViews: SystemViewSet, aShapes: List[Array[Int]], cShapes: List[Array[Int]]
){

  private val k = aShapes.head.sum // only used in assertions, I think

  type View = Views.View

  private type SubSystemView = ExtenderMap.SubSystemView

  /** The store holding information about system views so far.  Each system view
    * is represented in the domain as its server states and component view.
    * Each entry (servers, view) -> states in store(f) means that for each
    * state in states, a system view equivalent to (servers, view+state) has
    * been seen; further, state has component process identity type f. */
  private val store: ExtenderMap = new ShardedExtenderMap

  /** The number of SubSystemViews in the store. */
  def size = store.size 

  /** The number of States in the range of store.  */
  def stateCount: Int = store.stateCount


  /** All ways of splitting off one component from v.
    * @return a list of all (v1,st) such that v = v1 union st. */
  private def splits1(v: View): List[(View, State)] = {
    if(k == 1) List( (Array[State](), v(0)) )
    else if(k == 2){
      val st1 = v(0); val st2 = v(1)
      List( (Array(st2), st1), (Array(st1), st2) )
    }
    else
      ( for(i <- 0 until k) yield (v.patch(i, Array[State](), 1), v(i)) ).toList
  }

  /** Add sv to the store. */
  private def add(sv: SystemView) = {
    val servers = sv.serverStates; val v = sv.componentView
    assert(v.length == k)
    for((v1, st) <- splits1(v)){
      val (rv, rsts) = Remapper.remapSplitCanonical(v1, st, sv.rhoS)
      if(rsts.nonEmpty){
        val f = st.family; val ssv = new SubSystemView(servers, rv)
        store.add(f, ssv, rsts)
      }
    }
  }  

  @inline private def insert(cpts: View, st: State): View = {
    val cs = st.cs; var i = 0
    assert(cpts.length == k)
    var result = new Array[State](k+1)
    while(i < k && cpts(i).cs < cs){ result(i) = cpts(i); i += 1 }
    result(i) = st
    while(i < k){ result(i+1) = cpts(i); i += 1 }
    result
  }

  /** All concretizations formed by extending sv. */
  private def extensions(sv: SystemView): List[SystemView] = { 
    val servers = sv.servers; val v = sv.componentView
    assert(v.length == k)
    val vShape = Views.shape(v)  // shape of v (by component families)
    var result = List[SystemView]() // result list

    for(f <- 0 until numFamilies){
      // Consider extending v with an element of process type f.
      vShape(f) += 1                                  // (*)
      if(cShapes.exists((sh: Array[Int]) => sh.sameElements(vShape))){
        // The intersection of all candidates so far
        var intersection: ArrayBuffer[State] = null 
        // Bit map showing which control states are possible; null means all
        var splits = splits1(v); assert(splits.nonEmpty)
        while(splits.nonEmpty){
          val (w, st) = splits.head; splits = splits.tail
          // Does aShapes contain shapes corresponding to w with a state of
          // type t added?  Set vShape to this shape
          val stType = st.family; vShape(stType) -= 1   // (**)
          if(aShapes.exists((sh: Array[Int]) => sh.sameElements(vShape))){
            // Get candidates for this split
            val theseCandidates = candidates(sv, w, st, f)
            if(intersection == null) intersection = theseCandidates
            else intersection = intersection.intersect(theseCandidates)
          }
          vShape(stType) += 1                       // undo update at (**)
        } // end of while(splits.nonEmpty)
      
        vShape(f) -= 1                                // undo update at (*)
        assert(intersection.nonEmpty, v+"; f = "+f)

        // Now insert each state into addSt
        for(addSt <- intersection)
          result ::= SystemView.mkSystemView1(servers, insert(v, addSt), sv.rhoS)
      } // end of if(cShapes.exists...)
    } // end of for(t <- 0 until numTypes)

    result
  }

  /** Given a split (v1,st) of a View v from sv, return extensions of v1
    * compatible with the store, i.e. states addSt such that a SystemView
    * equivalent to (servers, v1+addSt) was previously added, and the identity
    * type of addSt is extendProcType.  (One representative is returned for each
    * equivalence class.) 
    * @param allowedCS bit map showing which control states are allowed, with 
    * null meaning all are allowed. */
  private def candidates(
    sv: SystemView, v1: View, st: State, extendProcType: Family)
      : ArrayBuffer[State] = {
    val (rv, map1, freeIds) = Remapper.remapSplit1Canonical(v1, st, sv.rhoS)
    // rv is the remapped version of v1; map1 shows how to remap parameters of
    // rv back to v1; freeIds is the new identities in sts, not in v1 or rhoS.
    // if(verbose) println("remapped v1 to "+showView(rv)+"\n"+
    //                       "freeIds = "+freeIds.mkString(", "))
    // Find those extensions of rmv
    val ssv = new SubSystemView(sv.serverStates, rv)
    val addStates: Iterator[State] = store.get(extendProcType, ssv)
    assert(addStates.nonEmpty)
    // assert(addStates.forall(st1 => st1.idType == extendType)) 
    // if(verbose) println("states to add:"+addStates.mkString("\t"))

    // Rename each element of addStates by map1, mapping each new
    // parameter both to itself and to each identity in freeIds, but
    // not creating a state with process identity equal to stId.
    val stId: Parameter = st.componentProcessIdentity 
    var results = new ArrayBuffer[State]
    for(addSt <- addStates)
      allRenamings(addSt, map1, freeIds, sv.numParams, stId, results)
    // No need to clone sv.numParams, because it is newly created
    results
  }

  /** Rename addState according to map, but mapping each other (new)
    * variable (t,id) (where id >= Remapper.SplitFreshVar) both to
    * nextId(t) and each identity in freeIds(t), but not mapping the
    * identity variable to notId.  Add results to buffer.
    * 
    * Note: freeIds and nextIds are temporarily updated, but returned to their
    * original state. */
  private def allRenamings(
    addState: State, map: Array[Array[Int]], freeIds: Array[List[Identity]],
    nextId: Array[Identity], notProcId: ComponentProcessIdentity, 
    buffer: ArrayBuffer[State]) 
      : Unit = {
    val (fNotId,notIdId) = notProcId
    // assert(nextId(indexingTypes(fNotId)) > notIdId)
    val family = addState.family; val typeMap = addState.typeMap
    val ids = addState.ids; val id = ids(0)

    if(fNotId != family || id >= Remapper.SplitFreshVal || 
         map(typeMap(0))(id) != notIdId){
      // Map showing how new variables have been renamed so far.
      val mapNew = Remapper.newRemappingMap
      // An entry mapNew(t)(i) = arg >= 0 means new parameter (t,
      // Remapper.SplitFreshVal+i) has been remapped to (t,arg).  Note: mapNew
      // is temporarily updated within allRenamings0, but restored to its
      // previous value.
      // Buffer to hold the renamed parameters
      val renamings = new ArrayBuffer[Array[Identity]]
      val cs = addState.cs; val n = ids.length
      allRenamings0(0) // Rename the parameters
      // Create resulting States
      for(ps1 <- renamings) buffer += State(family, cs, ps1)

      // Add id in position ix of each element of renamings[start..).  This
      // has the effect of prefixing every such partial renaming with id.
      @inline def prefix(id: Identity, ix: Int, start: Int) =
        for(i <- start until renamings.length) renamings(i)(ix) = id

      // We perform the renaming from right to left.  Rename ids from ix
      // onwards.  Put all resulting renamings in renamings.  Note we need
      // this rather convoluted structure, because the temporary updates to
      // freeIds and rho have to be carried forward through the remainder of
      // the renaming.
      def allRenamings0(ix: Int): Unit = {
        if(ix == n) renamings += State.getIdentityArray(n) // new Array[Identity](n)
        else{
          val start = renamings.length; val arg = ids(ix); val t = typeMap(ix)
          if(isDistinguished(arg)){ // don't rename arg
            allRenamings0(ix+1); prefix(arg, ix, start)
          }
          else if(arg < Remapper.SplitFreshVal){
            val arg1 = map(t)(arg)
            // assert(arg1 >= 0,
            //        "\nmap = "+map.map(_.mkString(", ")).mkString("; ")+"\n"+
            //          (t,arg)+"; "+ix)
            allRenamings0(ix+1); prefix(arg1, ix, start)
          }
          else{ // new parameter
            val argShift = arg-Remapper.SplitFreshVal;
            val arg1 = mapNew(t)(argShift)
            if(arg1 >= 0){ allRenamings0(ix+1); prefix(arg1, ix, start) }
            else{ // Need to rename new parameter
              // Option 1: map to nextId(t)
              val arg2 = nextId(t); nextId(t) += 1 // temporary update (*)
              if(arg2 >= typeSizes(t)){
                println("Not enough parameters of type "+typeNames(t)+
                          " in script to produce renaming")
                // println(addState.toString0); really need whole view
                sys.exit
              }
              mapNew(t)(argShift) = arg2           // temporary update (*)
              allRenamings0(ix+1); prefix(arg2, ix, start)
              nextId(t) -= 1; mapNew(t)(argShift) = -1 //restore values from (*))

              // Option 2: map arg to each value id in freeIds(t)
              val freeIdsT = freeIds(t) // store this
              for(id <- freeIdsT; if ix != 0 || family != fNotId || id != notIdId){
                freeIds(t) = freeIdsT.filter(_ != id) // temporary update
                mapNew(t)(argShift) = id       // temporarily update mapNew (**)
                val start1 = renamings.length
                allRenamings0(ix+1); prefix(id, ix, start1)
                mapNew(t)(argShift) = -1       // undo update to mapNew at (**)
              }
              freeIds(t) = freeIdsT // restore previous value
            } // end of else (need to rename new parameter)
          } // end of else (new parameter)
        }
      } // end of allRenamings0
    } // end of if(fNotId != family || ...)
  } // end of allRenamings

  /** Number of new views created. */
  val viewCount = new AtomicInteger(0)

  private val Million = 1000000

  /** The gamma function applied to newViews.  Each new view found is added to
    * concViews. */
  def gamma(newViews: Array[SystemView]): Array[SystemView] = {
    val count = new AtomicInteger(0); viewCount.set(0)
    print("NewViewExtender.gamma: #newViews = "+printInt(newViews.length)+": ")
    for(sv <- newViews.par) add(sv)
    val newConcViews = new ShardedConcBuffer[SystemView] // new views found
    for(sv <- newViews.par){
      val c = count.getAndIncrement // # system views expanded
      if((c&31) == 0 && c%100000 == 0){ 
        if(c%Million == 0) print(c/Million+"M") else print(".") 
      }
      gamma1(sv, newConcViews)
    }
    println
    newConcViews.get
  }

  /** The gamma function applied to sv.  Each view found is added to
    * concViews, and new views found added to newConcViews. */
  @inline private
  def gamma1(sv: SystemView, newConcViews: ConcBuffer[SystemView]) = {
    for(conc <- extensions(sv)){
      if(concViews.add(conc)){
        // system.checkTypeSizes(conc) -- now done in State
        newConcViews.put(conc)
        val vc = viewCount.incrementAndGet // # system views produced
        if((vc&63) == 0 && vc%Million == 0) print("<"+vc/Million+"M>")
      }
    }
  }
  // ~90% of the time of the above function is spent in extensions.
}

// -------------------------------------------------------

// Following now in ExtenderMap
// object NewViewExtender{
//   type View = Views.View

//   /** A subview of a system view; used in the domain of the store mapping. */
//   class SubSystemView(val servers: ServerStates, val view: View){
//     override def equals(that: Any) = {
//       val ssv = that.asInstanceOf[SubSystemView]
//       ssv.servers == servers && ssv.view.sameElements(view)
//     }

//     override def hashCode = {
//       var h = servers.hashCode
//       for(st <- view) h = h*73 + st.hashCode
//       h
//     }

//     override def toString =
//       "SubSystemView("+servers.servers+", "+view.mkString("<", ",", ">")+")"
//   } // end of SubSystemView
// }
