package ViewAbstraction

// import ox.cads.util.Profiler

import scala.math.Ordering.Implicits._ // for sorting list of lists

/** Utility object to apply remapping of parameters in states.  
  * 
  * Some remapping functions take a ParamMap, showing which parameters should
  * be left unchanged.  Each function remaps parameters so that they, and
  * those in the ParamMap if applicable, form an initial segment of the
  * naturals, and their first occurences are in order. */ 
object Remapper{
  type View = Views.View

  /** Max number of values of each type that we need to keep track of in any
    * remapping. */
  val rowSizes = Array.tabulate(numTypes)( t => 
    typeSizes(t) + State.maxParamsOfType(t))
  println("Remapper.rowSizes = "+rowSizes.mkString(", "))

  /** The type of maps recording the values that parameters get remapped to.  
    * map(t)(arg) records the value that arg of type t gets remapped to,
    * or -1 if no value has been allocated yet.  I.e., it represents the
    * mapping
    *  {(t,arg) -> (t, map(t)(arg)) |
    *       0 <= t < numTypes, 0 <= arg < size(t), map(t)(arg) != -1}.
    */
  private type RemappingMap = Array[Array[Int]]

  /** A clear RemappingMap, representing the empty mapping; used as a
    * template for creating more such. */
  private val remappingMapTemplate: RemappingMap = 
    Array.tabulate(numTypes)(t => Array.fill(rowSizes(t))(-1))

  /** A thread-local RemappingMap. */
  private object ThreadLocalRemappingMap extends ThreadLocal[RemappingMap]{
    override def initialValue() = {
      Array.tabulate(numTypes)(t => Array.fill(rowSizes(t))(-1))
      // val map = new Array[Array[Int]](numTypes); var t = 0
      // do{ map(t) = remappingMapTemplate(t).clone; t += 1 } while(t < numTypes)
      // assert(map.forall(_.forall(_ == -1)))
      // map
    }
  }

  /** A clear RemappingMap, representing the empty mapping.  Different
    * calls to this will use the same arrays, so two RemappingMaps
    * created by the same thread should not be in use at the same time. */
  @inline def newRemappingMap: RemappingMap = {
    // Profiler.count("newRemappingMap")
    val map = ThreadLocalRemappingMap.get; var t = 0
    do{
      var i = 0; val len = map(t).length
      do{ map(t)(i) = -1; i += 1 } while(i < len)
      t += 1
    } while(t < numTypes)
    //assert(map.forall(_.forall(_ == -1)))
    map 
  }

  // -----

  // /** A clear RemappingMap, representing the empty mapping.  This
  //   * version does not share arrays with other RemappingMaps. */
  // @inline private def newRemappingMapUnique: RemappingMap = {
  //   val map = ThreadLocalRemappingMapUnique.get 
  //    // new Array[Array[Int]](numTypes);
  //   var t = 0
  //   do{ map(t) = remappingMapTemplate(t).clone; t += 1 } while(t < numTypes)
  //   assert(map.forall(_.forall(_ == -1)))
  //   map
  // }

  /** Create a RemappingMap corresponding to rho, i.e. the identity map
    * on (t,i) for i <- [0..rho(t), for each t. */
  private def createMap(rho: ParamMap): RemappingMap = {
    val map = newRemappingMap
    for(t <- 0 until numTypes){
      val len = rho(t).length; var i = 0
      while(i < len){ map(t)(i) = i; i += 1 }
      // while(i < rowSizes(t)){ assert(map(t)(i) == -1); i += 1 }
    }
    map
  }

  /** The type of maps giving the next argument to map a parameter of
    * type t.  The corresponding RemappingMap has domain all
    * parameters (t,i) for i <- [0..nextArg(t)), for each t. */
  private type NextArgMap = Array[Int]

  /** A thread-local NextArgMap. */
  private object ThreadLocalNextArgMap extends ThreadLocal[NextArgMap]{
    override def initialValue() = new Array[Int](numTypes)
  }

  /** Create a new NextArgMap, corresponding to the empty mapping. */
  @inline private def newNextArgMap: NextArgMap = {
    // We re-use previous maps, to reduce GC churn
    val map = ThreadLocalNextArgMap.get
    for(t <- 0 until numTypes) map(t) = 0
    map
    // new Array[Int](numTypes)
  }

  /** Create a new NextArgMap corresponding to rho. */
  @inline private def createNextArgMap(rho: ParamMap): NextArgMap = {
    val naMap = ThreadLocalNextArgMap.get // new Array[Int](numTypes)
    for(i <- 0 until numTypes) naMap(i) = rho(i).length
    naMap
    // rho.map(_.length)
  }

  // ======================================================= Helper functions

  /** Remap the parameter (t,arg), updating map and arg appropriately. */
  @inline private def remapArg(
    map: RemappingMap, nextArg: NextArgMap, t: Type, arg: Identity)
      : Identity = {
    if(isDistinguished(arg)) arg
    else{
      // assert(arg < SplitFreshVal, arg+": "+map(t).mkString(", ")) 
      val arg1 = map(t)(arg)
      if(arg1 < 0){
        map(t)(arg) = nextArg(t); nextArg(t) += 1; nextArg(t)-1
      }
      else arg1
    }
  }

  /** Remap a single state st, updating map and nextArg.
    * @param mkNew should a new (temporary) State be created, not stored in the 
    * StateMap? */
  @inline private def remapState(
    map: RemappingMap, nextArg: NextArgMap, st: State)
      : State = {
    val cs = st.cs; val ids = st.ids
    val len = ids.length; var index = 0
    val remappedParams = State.getIdentityArray(len) // new Array[Int](len)
    val tArray = State.stateTypeMap(cs) // array giving types of ids
    while(index < len){
      remappedParams(index) = remapArg(map, nextArg, tArray(index), ids(index))
      index += 1
    }
    // if(mkNew) new State(st.family, cs, remappedParams)   // else 
    State(st.family, cs, remappedParams)
  }

  /** Remap procs, updating map and nextArg. 
    * @param mkNew should a new (temporary) State be created, not stored in the 
    * StateMap? */
  @inline private def remapStates(
    map: RemappingMap, nextArg: NextArgMap, procs: List[State]): List[State] = {
    procs.map(st => remapState(map, nextArg, st)) // , mkNew
  }

  /** Remap v, updating map and nextArg. */
  @inline private def remapView
    (map: RemappingMap, nextArg: NextArgMap, v: View): View = {
    val len = v.length
    val newView = new Array[State](len) // IMPROVE: recycle
    for(i <- 0 until len) newView(i) = remapState(map, nextArg, v(i))
    newView
  }


  // =================================== Permuting states

  /** All permutations of v that preserve that control states are in increasing
    * order. */
  @inline private def viewPerms(v: View): List[List[State]] = {
    // assert(v.nonEmpty) 
    viewPermsFrom(v, 0)
  }

  /** All permutations of v[from..) ordered by control states. */
  private def viewPermsFrom(v: View, from: Int): List[List[State]] = {
    // We avoid recursing on the empty sub-view
    val len = v.length
    // Profiler.count("Remapper.viewPermsFrom"+(len-from))
    if(from == len-1) List(List(v(from)))
    else{
      val h = v(from); val cs = h.cs; var i = from+1
      while(i < len && v(i).cs == cs) i += 1
      // States v[from..i) all have the same control state.
      if(i == from+1) viewPermsFrom(v, i).map(h :: _) // optimise for this case
      else if(i == from+2){ // and this case
        val second = v(from+1)
        if(i == len) List(List(h,second), List(second,h))
        else{ 
          val tails = viewPermsFrom(v, i)
          var result = List[List[State]]()
          for(tail <- tails)
            result = (h::second::tail) :: (second::h::tail) :: result
          result
        }
      }
      else if(i == from+3){ // and this case
        val snd = v(from+1); val thd = v(from+2)
        if(i == len) 
          List( List(h,snd,thd), List(h,thd,snd), List(snd,h,thd),
                 List(snd,thd,h), List(thd,h,snd), List(thd,snd,h) )
        else{ 
          val tails = viewPermsFrom(v, i)
          var result = List[List[State]]()
          for(tail <- tails)
            result = (h::snd::thd::tail) :: (h::thd::snd::tail) :: 
              (snd::h::thd::tail) :: (snd::thd::h::tail) ::
              (thd::h::snd::tail) :: (thd::snd::h::tail) :: result
          result
        }
      }
      else{
        val inits = myPerms1(v, from, i)
        if(i == len) inits // optimise for this case
        else{
          val tails = viewPermsFrom(v, i)
          var result = List[List[State]]()
          for(init <- inits; tail <- tails) result ::= append1(init, tail)
          result
        }
      }
    }
  }

  /** All permutations of v[from..end).  Pre: this has length at least 2. */
  @inline private def myPerms1(v: View, from: Int, end: Int)
      : List[List[State]] = {
    val len = end-from
    // Profiler.count("Remapper.myPerms1:"+len)
    require(len >= 3)
    if(len == 3){ // Optimise for this case
      val st1 = v(from); val st2 = v(from+1); val st3 = v(from+2)
      List( List(st1,st2,st3), List(st1,st3,st2), List(st2,st1,st3), 
             List(st2,st3,st1), List(st3,st1,st2), List(st3,st2,st1) )
    }
    // if(len == 2){ // Optimise for this case
    //   val st1 = v(from); val st2 = v(from+1)
    //   List( List(st1,st2), List(st2,st1) )
    // }
    else{
      val xs = new Array[State](len) // IMPROVE: can we avoid this?
      for(i <- 0 until len) xs(i) = v(i+from)
      // val xs = Array.tabulate(end-from)(i => v(i+from))
      myPermsA(xs, len)
      // var stateList = List(v(end-3), v(end-2), v(end-1)); var i = end-4
      // while(i >= from){ stateList ::= v(i); i -= 1 }
      // myPerms(stateList)
    }
  }

  /** All permutations of xs. */
  @inline private def myPerms[A](xs: List[A]): List[List[A]] = {
    val len = xs.length // ; assert(len >= 2)
    // if(len == 1) List(xs)
    // else 
    if(len == 2){
      val x1::x2::_ = xs; List( List(x1,x2), List(x2,x1) )
    }
    else
      (0 until len).toList.flatMap{ i =>
        val (before, x::rest) = xs.splitAt(i)
        myPerms(append1(before,rest)).map(x::_)
      }
  }

  /** All permutations of xs[0..len). */
  @inline private def myPermsA[A](xs: Array[A], len: Int): List[List[A]] = {
    // assert(len >= 2)
    // if(len == 1) List(List(xs(0)))
    // else 
    if(len == 2){
      val x0 = xs(0); val x1 = xs(1); List(List(x0,x1), List(x1,x0))
    }
    else (0 until len).toList.flatMap{ i =>
      val x = xs(i); xs(i) = xs(len-1)
      val ps0 = myPermsA(xs, len-1); xs(i) = x
      ps0.map(x::_)
    }
  }

  /** Convert components to canonical form, respecting rhoS.  
    * Pre: control states of components are ordered. */
  def mkCanonical1(components: View, rhoS: ParamMap): (View, Array[Int]) = {
    // assert(components.forall(st => st.ids.forall(_ < SplitFreshVal)))
    // println("mkCanonical1: "+
    //           components.map(_.toString0).mkString("<", ", ", ">"))
    mkCanonicalHelper(components, createMap(rhoS), createNextArgMap(rhoS))
  }

  /** Convert servers and components into canonical form.
    * @return the canonical form for servers and components, and the number of 
    * parameters of each type. */
  def mkCanonical2(servers: List[State], components: View)
      : (List[State], View, Array[Int]) = {
    // clear
    val map = newRemappingMap; var nextArg = newNextArgMap
    val newServers = remapStates(map, nextArg, servers) // , false)
    val (newCpts, newNextArg) = mkCanonicalHelper(components, map, nextArg)
    (newServers, newCpts, newNextArg) 
  }

  /** Convert components into canonical form, remapping according to map
    * and nextArg. */
  private def mkCanonicalHelper(
    components: View, map: RemappingMap, nextArg: NextArgMap)
      : (View, Array[Int]) = {
    // Profiler.count("Remapper.mkCanonicalHelper")
    val perms0: List[List[State]] = viewPerms(components)
    if(perms0.length == 1) // optimise for this case
      (remapView(map, nextArg, components), nextArg)
    else{
      // Variables to store clones of map and nextArg, to be used with each
      // remap of component states, below.  
      val newMap = map.map(_.clone) 
      var newNextArg: NextArgMap = nextArg.clone
      var minPerm: List[State] = 
        remapStates(newMap, newNextArg, perms0.head) 
      for(v <- perms0.tail){
        for(i <- 0 until numTypes) newMap(i) = map(i).clone
        newNextArg = nextArg.clone
        val nextPerm = remapStates(newMap, newNextArg, v)
        if(nextPerm < minPerm) minPerm = nextPerm 
      }
      (minPerm.toArray, newNextArg) // IMPROVE: use recycled array
    }
  }


  // =================================== Remapping of splits, in two directions.
  // Used in NewViewExtender

  val SplitFreshVal = State.SplitFreshVal

  /** Remap v0 to its canonical form, and then remap st, mapping new parameters
    * in the latter to large values, at least SplitFreshVal. */
  def remapSplitCanonical(v0: View, st: State, rhoS: ParamMap)
      : (View, List[State]) = {
    // IMPROVE: it might be better to remap st only for the canonical v1,
    // which means saving the corresponding map.
    val allSplits = for(v <- viewPerms(v0)) yield{
      val map = createMap(rhoS); val nextArg = createNextArgMap(rhoS)
      // reset(rhoS)
      val v1 = remapStates(map, nextArg, v) // , false) 
      for(i <- 0 until numTypes) nextArg(i) = SplitFreshVal
      (v1, remapState(map, nextArg, st)) // , false))
    }
    val v1 = allSplits.map(_._1).min
    val remappedSts = (for((`v1`, sts1) <- allSplits) yield sts1).distinct
    if(remappedSts.length > 1) assert(v0.length > 1)
    (v1.toArray, remappedSts)
  }

  /** A thread-local RemappingMap for use in remapSplit1Canonical. */
  private object ThreadLocalSplitRemappingMaps
      extends ThreadLocal[(RemappingMap, Array[List[Identity]])]{
    override def initialValue() = {
      (new Array[Array[Int]](numTypes), new Array[List[Identity]](numTypes))
      // val map = new Array[Array[Int]](numTypes); var t = 0
      // do{ map(t) = remappingMapTemplate(t).clone; t += 1 } while(t < numTypes)
      //   assert(map.forall(_.forall(_ == -1)))
      // map
    }
  }

  /** Remap v0 to its canonical form; also return ParamMap for variables in
    * remapped v0, and parameters that are new in st (i.e. not in rhoS or v0).
    * @return a tuple (rv, map1, newArgs) where: rv is the remapped
    * version of v0; map1 is a map to rename rv back to v0; newArgs gives the
    * new parameters in sts that didn't appear in v0 or rhoS. */
  def remapSplit1Canonical(v0: View, st: State, rhoS: ParamMap)
      : (View, Array[Array[Int]], Array[List[Identity]]) = {
    // min value found so far, and corresponding maps
    var v1Min: List[State] = null 
    // Map to rename v1Min back to v, and list of fresh parameters.  Note: the
    // former exists concurrently with the ThreadLocalRemappingMap, so needs
    // to be distinct.
    val (map1, newArgs) = ThreadLocalSplitRemappingMaps.get
    // val map1: RemappingMap =  ThreadLocalRemappingMapUnique.get
    // val newArgs = new Array[List[Identity]](numTypes) 
    for(v <- viewPerms(v0)){
      val map = createMap(rhoS); val nextArg = createNextArgMap(rhoS)
      val v1 = remapStates(map, nextArg, v)
      if(v1Min == null || v1 < v1Min){
        // This is a new min
        v1Min = v1
        // Create map to rename v1 back to v
        for(t <- 0 until numTypes){
          map1(t) = remappingMapTemplate(t).clone
          for(i <- 0 until rowSizes(t)){
            val j = map(t)(i); if(j >= 0) map1(t)(j) = i
          }
        }
        // map1 = newRemappingMapUnique
        // for(t <- 0 until numTypes; i <- 0 until rowSizes(t)){
        //   val j = map(t)(i); if(j >= 0) map1(t)(j) = i
        // }
        // Find parameters of sts not in v0 or rhoS, and hence not in map at this
        // point
        // newArgs = Array.fill(numTypes)(List[Identity]())
        for(t <- 0 until numTypes) newArgs(t) = List[Identity]()
        val ids = st.ids; var index = 0; val n = ids.length
        while(index < n){
          val id = ids(index); val t = st.typeMap(index); index += 1
          if(id >= 0 && map(t)(id) == -1 && !newArgs(t).contains(id))
            newArgs(t) ::= id
        }
      }
    } // end of for
    (v1Min.toArray, map1, newArgs)

    // val allSplits = for(v <- viewPerms(v0)) yield{
    //   val map = createMap(rhoS); val nextArg = createNextArgMap(rhoS)
    //   val v1 = remapStates(map, nextArg, v) 
    //   // Create map to rename v1 back to v; note: this exists
    //   // concurrently with other RemappingMaps, so needs to be unique.
    //   // IMPROVE: can this be thread-local, but distinct from those produced by
    //   // newRemappingMap?
    //   val map1 = newRemappingMapUnique 
    //   for(t <- 0 until numTypes; i <- 0 until rowSizes(t)){
    //     val j = map(t)(i); if(j >= 0) map1(t)(j) = i
    //   }
    //   // Find parameters of sts not in v0 or rhoS, and hence not in map at this
    //   // point
    //   val newArgs = Array.fill(numTypes)(List[Identity]())
    //   val cs = st.cs; val ids = st.ids
    //   var index = 0; val n = ids.length
    //   while(index < n){
    //     val id = ids(index)
    //     val t = State.stateTypeMap(cs)(index)
    //     index += 1
    //     if(id >= 0 && map(t)(id) == -1 && !newArgs(t).contains(id))
    //       newArgs(t) ::= id
    //   }
    //   // for(id <- st.ids){
    //   //   val t = stateTypeMap(st.cs)(index)
    //   //   index += 1
    //   //   if(id >= 0 && map(t)(id) == -1 && !newArgs(t).contains(id))
    //   //     newArgs(t) ::= id
    //   // }
    //   (v1, map1, newArgs)
    // }
    // val (v1, map1, newArgs) = allSplits.minBy(_._1)
    // (v1.toArray, map1, newArgs) 
  }
}
