package ViewAbstraction

// import ox.cads.util.Profiler

import scala.collection.mutable.ArrayBuffer

/** States of processes. 
  * @param family the family of components that this belongs to, or -1 for a 
  * server.
  * @param cs the control state of this state.
  * @ids the identity parameters of this state.
  * @param isNew is this state being created during the compilation stage, 
  * in which case checking the type sizes are big enough is not necessary. */
class State(val family: Family, val cs: ControlState, 
            val ids: Array[Identity], isNew: Boolean = false){

  /** The component process identity. */
  def componentProcessIdentity: ComponentProcessIdentity = {
    // assert(family >= 0)
    (family, ids.head)
  }

  /** The types of parameters. */
  def typeMap: Array[Type] = State.stateTypeMap(cs)

  /** Check the type sizes from the script are large enough for all the
    * parameters in this State.  This is not done during compilation, because
    * typeMap is not yet initialised, and the State is bound to pass the test,
    * anyway. */
  @inline private def checkTypeSizes() = {
    val tm = typeMap
    for(i <- 0 until ids.length){
      if(ids(i) < State.SplitFreshVal && ids(i) >= typeSizes(tm(i))) 
        synchronized{
          val t = tm(i)
          println("Not enough identities of type "+typeNames(t)+" in script ("+
                    typeSizes(t)+") to represent system view\n"+toString0)
          sys.exit
        }
    }
  }

  if(!isNew) checkTypeSizes()

  // Equality test.
  // Equality is now reference equality, as we avoid creating duplicate states.
  // @inline override def equals(that: Any) = that match{
  //   case st: State => (st eq this) || st.cs == cs && st.ids.sameElements(ids)
  // }
  // assert(-1 <= family && family < 2) 
  assert(family == -1 || ids.nonEmpty) // components have an identity

  override val hashCode = mkHash

  /** Hash function over states. */
  @inline private def mkHash: Int = {
    var h = cs*41+family; var i = 0; var n = ids.length
    while(i < n){ h = h*41+ids(i); i += 1 }
    h
  } 

  /** Convert this to a String.  This version assumes that all the parameters
    * correspond to values in the script. */
  override def toString = {
    val paramsString = (0 until ids.length).map{ j =>
      val t = State.stateTypeMap(cs)(j); scriptNames(t)(ids(j)) }
    cs+"["+family+"]"+paramsString.mkString("(", ",", ")")
  }

  /** Convert this to a String.  If this uses a parameter not in the script, it
    * is represented by its raw (type, value) representation. */
  def toString0 = {
    val paramsString = (0 until ids.length).map{ j =>
      val t = State.stateTypeMap(cs)(j)
      if(scriptNames == null) "*"+(t,ids(j))+"*"
      else scriptNames(t).get(ids(j)) match{
        case Some(st) => st
        case None => "*"+(t,ids(j))+"*"
      }
    }
    cs+"["+family+"]"+paramsString.mkString("(", ",", ")")
  }  

  def toString00 = cs+"["+family+"]"+ids.mkString("(", ",", ")")

  /** the type of the identity. */
  // private lazy val idType = State.idType(cs) // State.stateTypeMap(cs)(0)

  /** The identity parameter. */
  // lazy val identity: Parameter = (State.stateTypeMap(cs)(0), ids.head)


} // End of State class

// ------------------------------------------------ Companion object

object State{
  /** Mapping storing all the States found so far. */
  private var stateStore: MyStateMap = new MyStateHashMap

  /* We initially use a MyStateHashMap, but later transfer values into a
   * MyTrieStateMap (in renewStateStore).  This is because the latter
   * makes assumptions about various data structures being
   * initialised, but some States are created before this happens. */

  /** The number of States stored. */
  def stateCount = stateStore.size

  /** Factory method to either find existing state matching (family, cs,
    * ids), or to create new one. */
  @inline def apply(
    family: Int, cs: ControlState, ids: Array[Identity], isNew: Boolean = false)
      : State = {
    // Profiler.count("State")
    stateStore.getOrAdd(family, cs, ids, isNew)
    // assert(st.family == family)
  }

  @inline def register(st: State): State = {
    // Profiler.count("State.register")
    stateStore.add(st)
  }

  /** Replace the initial StateMap with a Trie-based one. */
  def renewStateStore(numCS: Int, minCS: Int) = /*if(false)*/{
    print("Creating Trie-based StateMap...")
    val it = stateStore.asInstanceOf[MyStateHashMap].iterator
    val tsm = new MyTrieStateMap(numCS, minCS)
    for(st <- it) tsm.add(st) 
    assert(tsm.size == stateStore.size, tsm.size+"; "+stateStore.size)
    stateStore = tsm
    println(".  Done.")
  }

  /** Ordering over states. */
  implicit object StateOrdering extends Ordering[State]{
    /** Ordering on states. */
    def compare(st1: State, st2: State) = {
      if(st1.cs != st2.cs) st1.cs - st2.cs
      else{
        var i = 0; var n = st1.ids.length // assert(st2.ids.length == n)
        while(i < n && st1.ids(i) == st2.ids(i)) i += 1
        if(i == n) 0 else st1.ids(i) - st2.ids(i)
      }
    }
  }

  // ----- Memory management functions

  /** Supply of Array[Identity]s, to prevent GC churning.  
    * Entry len contains a supply of Array[Identity]s of length len.  */
  private object ThreadLocalIdentityArraySupply
      extends ThreadLocal[Array[List[Array[Identity]]]]{
    // Initialise to size 4, as this is normally enough
    override def initialValue() = Array.fill(4)(List[Array[Identity]]())
  }

  /** Get an Array[Identity] of size len, reusing a previous one if possible. */
  @inline def getIdentityArray(len: Int): Array[Identity] = /*if(false)*/{
    val ab = ThreadLocalIdentityArraySupply.get
    if(len < ab.length){
      val list = ab(len)
      if(list.nonEmpty){
        //Profiler.count("get success")
        ab(len) = list.tail; list.head
      }
      else{ /* Profiler.count("get fail"); */ new Array[Identity](len) }
    }
    else{ /* Profiler.count("get fail"); */ new Array[Identity](len) }
  }
  // else new Array[Identity](len)

  /** Return ids for recycling. */
  @inline def returnIdentityArray(ids: Array[Identity]) = /*if(false)*/{
    val ab = ThreadLocalIdentityArraySupply.get; val len = ids.length
    // Expand to size len if necessary
    if(ab.length <= len){
      // println("expanding")
      val ab1 = new Array[List[Array[Identity]]](len+1)
      for(i <- 0 until ab.length) ab1(i) = ab(i)
      for(i <- ab.length until len) ab1(i) = List[Array[Identity]]()
      ab1(len) = List(ids)
      ThreadLocalIdentityArraySupply.set(ab1)
    }
    else ab(len) ::= ids
    // Profiler.count("return"+ab(len).length/10)
  }

  // ----- Variables and functions concerning states. 

  /** Minimum value that new identities are mapped to, when splitting
    * views in Remapper.remapSplitCanonical. */
  val SplitFreshVal = 1 << 30

  /** Array storing information about types of parameters.
    * stateTypeMapArray(cs-minCS) stores information about the types of
    * parameters for control state cs.  */
  private var stateTypeMapArray: Array[Array[Type]] = null

  /** Displacement within stateTypeMapArray. */
  var minCS = Int.MaxValue

  /** Number of control states. */
  var numCS = -1

  /** Array giving the maximum number of parameters of any type in a state. */
  var maxParamsOfType: Array[Int] = null

  /** Array giving the type of the identity for control states.
    * idTypeArray(cs-minCS) gives the type of processes with control state cs,
    * or -1 for server processes with no identity.  */
  private var idTypeArray: Array[Type] = null

  /** Initialise the stateTypeMapArray, idTypeArray and minCS.  Also
    * update the StateMap.  Called by
    * FDRTransitionMap.createStateTypeMap, called by System.init. */
  def setStateTypeMap(stma: Array[Array[Type]], minCS: Int) = {
    stateTypeMapArray = stma; this.minCS = minCS; numCS = stma.size
    // set idTypeArray; need to be careful about server states with no
    // parameters
    idTypeArray = stma.map(a => if(a != null && a.nonEmpty) a(0) else -1)
    // The filter(_ != null) is necessary, because extra control states are
    // created in FDRSession.getTypeInt, but don't have types recorded.
    maxParamsOfType = Array.tabulate(numTypes)( t =>
      stma.filter(_ != null).map(_.count(_ == t)).max )
    println("maxParamsOfType = "+maxParamsOfType.mkString(", "))
    renewStateStore(stma.length, minCS)
  }

  /** Array giving the types of identities for control state cs. */
  @inline def stateTypeMap(cs: ControlState): Array[Type] =
    stateTypeMapArray(cs-minCS)

  /** Type of the identity of component with control state cs. */
  private def idType(cs: ControlState): Type = idTypeArray(cs-minCS)
}
