package ViewAbstraction
import scala.collection.mutable.{Set,HashSet,ArrayBuffer,Map}
// import scala.math.Ordering.Implicits._ // for sorting list of lists
// import ox.cads.util.Profiler

/** A view on a state of a system.
  * @param servers the states of the servers, modelled explicitly.
  * @param componentView a view of the states of the components. 
  * @param numParams the number of parameters of each type.
  * @param numParamsSummary a summary of the number of parameters of each type,
  * rolled into a single Int.
  * @param serverIdsP the identities of each type used by servers.
  * 
  * The arguments must satisfy the following condition (creating a datatype
  * invariant):
  * 
  * For each type t, the identities of type t within servers++components
  * are [0..numParams(t)), and the first occurences of these
  * identities are in increasing order.
  * 
  * Subclasses may impose additional constraints. */
abstract class SystemView(val serverStates: ServerStates)
  // val componentView : Views.View,
  // numParams1: Array[Int], 
  // numParamsSummary: Int)
{
  /** A view of the states of the components.
    *  Defined in subclasses. */
  def componentView: Views.View

  // ----- Main functions used during a check

  /** The number of parameters of each type.
    *  This assumes numParamsSummary is passed in. */
  // @inline def numParams: Array[Int] = {
  //   val nPs = new Array[Int](numTypes); var sum = numParamsSummary
  //   var i = 0
  //   while(i < numTypes){
  //     nPs(i) = sum & SystemView.Mask; sum = sum >> SystemView.Shift; i += 1
  //   }
  //   val nPsX = numParamsX
  //   assert(nPs.sameElements(nPsX), 
  //          this+"\n"+nPs.mkString("<", ",", ">")+"; "+nPsX.mkString("<", ",", ">"))
  //   nPs
  // }

  /** The number of parameters of each type. */
  @inline def numParams: Array[Int] = {
    val nPs = serverStates.numParams.clone
    for(st <- componentView){
      val typeMap = st.typeMap // State.stateTypeMap(st.cs)
      var index = 0; val len = st.ids.length
      while(index < len){
        // val id = st.ids(index)
        val t = typeMap(index)
        nPs(t) = nPs(t) max (st.ids(index)+1); index += 1
      }
    }
    nPs
  }

  /** The list of states of the servers. */
  @inline def servers: List[State] = serverStates.servers

  /** For each type t, the list of identities of type t used within servers. */
  @inline def serverIds: Array[List[Identity]] = serverStates.serverIds

  /** Identity mapping on identities held by servers, for each type. */
  def rhoS = serverStates.rhoS

  // ----- Debugging code

  // Check invariant.  This is moderately expensive, so disabled by default.
  if(SystemView.Debugging){
    SystemView.checkInvariant(servers, componentView, numParams)
    val componentControlStates: Array[ControlState] = componentView.map(_.cs)
    assert(componentControlStates.sorted.sameElements(componentControlStates))
  }

  /** Check invariant holds; called in FullyConnectedSystemView. */
  def checkInvariant =
    SystemView.checkInvariant(servers, componentView, numParams)

  /** Check that the identites are correctly remapped. */
  private def checkIdsRemapped = {
    val nextIds = new Array[Identity](numTypes)
    for(st <- servers++componentView){
      var index = 0
      for(id <- st.ids){
        val t = State.stateTypeMap(st.cs)(index)
        index += 1
        assert(id <= nextIds(t))
        if(id == nextIds(t)) nextIds(t) += 1
      }
    }
    for(t <- 0 until numTypes) assert(nextIds(t) == numParams(t))
  }

  if(SystemView.Debugging){
    for(t <- 0 until numTypes){
      val sIds = serverIds(t)
      assert(sIds == sIds.sorted && (sIds.isEmpty || sIds.length == sIds.max+1),
             sIds)
    }
    checkIdsRemapped
  }

  // ----- Debugger information

  /* If this is a concretization, then predecessor -eventOrPly-> this.  If this
   * is an abstraction, then this was produced by applying alpha to
   * predecessor on ply number eventOrPly.  This double-use is designed to
   * save memory.  */

  /** The predecessor of this state under the transition relation. */
  private var predecessor: SystemView = null

  /** The event that lead to this state.   So predecessor -event-> this.  */
  private var eventOrPly: EventInt = -1

  /** If this is a concretization, record that this was produced by a transition
    * p-e-> this.  If this is an abstraction, record that this was produced by
    * applying alpha to p on ply e. */
  def setPred(p: SystemView, e: EventInt) = { predecessor = p; eventOrPly = e }

  /** If this is a concretization, get (p,e) such that this was produced by a
    * transition p-e-> this.  If this is an abstraction, get (p,e) such that
    * this was produced by applying alpha to p on ply e. */
  def getPred = (predecessor, eventOrPly)

  // ----- Printing

  /** Convert this to a String.  This version assumes that all the parameters
    * correspond to values in the script. */
  override def toString = {
    val cptsString = componentView.mkString(" || ")
    if(servers.nonEmpty)
      servers.mkString("[ ", " || ", " ]") + " || [ "+ cptsString +" ]"
    else cptsString
  }

  /** Convert this to a String.  If this uses a parameter not in the script, it
    * is represented by its raw (type, value) representation. */
  def toString0 = {
    val cptsString = componentView.map(_.toString0).mkString(" || ")
    if(servers.nonEmpty)
      servers.map(_.toString0).mkString("[ ", " || ", " ]") +
        " || [ "+ cptsString +" ]"
    else cptsString
  }

  // ----- Equality and hash code

  /** Are two lists of states equal.  Pre: the two lists have the same
    * length. */
  @inline private def equalStateLists(states1: List[State], states2: List[State])
      : Boolean =
    states1.isEmpty ||
      states1.head == states2.head &&
      equalStateLists(states1.tail, states2.tail)

  /** Does that have components equal to this?  Defined in subclasses. */
  protected def equalCpts(that: SystemView): Boolean

  /** Equality. */
  override def equals(that: Any) = {
    val f = that.asInstanceOf[SystemView]
    // assert(hashCode == f.hashCode) // guaranteed by hash table
    serverStates == f.serverStates && equalCpts(f)
  }

  /** Create hash of components and h0.  Defined in subclasses. */
  protected def cptsHash(h0: Int): Int

  // Override hash function
  @inline override def hashCode: Int = cptsHash(serverStates.hashCode)
}

// =======================================================

class SystemViewArray(
  override val serverStates: ServerStates,
  val componentView: Views.View)
 // numParamsSummary: Int)
extends SystemView(serverStates /*, numParamsSummary*/){

  @inline def equalCpts(that: SystemView) =
    componentView.sameElements(that.componentView)

  @inline def cptsHash(h0: Int): Int = {
    var h = h0
    for(st <- componentView) h = h*73 + st.hashCode
    h
  }
}

// =======================================================

class SystemView4(
  override val serverStates: ServerStates, //  numParamsSummary: Int,
  protected val st0: State, protected val st1: State, 
  protected val st2: State, protected val st3: State)
extends SystemView(serverStates /*, numParamsSummary */){

  @inline def componentView = Array(st0, st1, st2, st3)

  @inline def equalCpts(that: SystemView) = {
    val sv = that.asInstanceOf[SystemView4]
    sv.st0 == st0 && sv.st1 == st1 && sv.st2 == st2 && sv.st3 == st3
  }

  @inline def cptsHash(h0: Int): Int = 
    (((h0*73+st0.hashCode)*73+st1.hashCode)*73+st2.hashCode)*73+st3.hashCode
}

// =======================================================

class SystemView5(
  override val serverStates: ServerStates, //  numParamsSummary: Int,
  protected val st0: State, protected val st1: State, protected val st2: State,
  protected val st3: State, protected val st4: State)
extends SystemView(serverStates /* , numParamsSummary*/ ){
  @inline def componentView = Array(st0, st1, st2, st3, st4)

  @inline def equalCpts(that: SystemView) = {
    val sv = that.asInstanceOf[SystemView5]
    sv.st0 == st0 && sv.st1 == st1 && sv.st2 == st2 &&
      sv.st3 == st3 && sv.st4 == st4
  }

  @inline def cptsHash(h0: Int): Int = 
    ((((h0*73+st0.hashCode)*73+st1.hashCode)*73+st2.hashCode)
       *73+st3.hashCode)*73+st4.hashCode
}

// =======================================================

class SystemView6(
  override val serverStates: ServerStates, // numParamsSummary: Int,
  protected val st0: State, protected val st1: State, protected val st2: State,
  protected val st3: State, protected val st4: State, protected val st5: State)
extends SystemView(serverStates /* , numParamsSummary */){
  @inline def componentView = Array(st0, st1, st2, st3, st4, st5)

  @inline def equalCpts(that: SystemView) = {
    val sv = that.asInstanceOf[SystemView6]
    sv.st0 == st0 && sv.st1 == st1 && sv.st2 == st2 &&
      sv.st3 == st3 && sv.st4 == st4 && sv.st5 == st5
  }

  @inline def cptsHash(h0: Int): Int = 
    (((((h0*73+st0.hashCode)*73+st1.hashCode)*73+st2.hashCode)
        *73+st3.hashCode)*73+st4.hashCode)*73+st5.hashCode
}


// ========================================================

/** Companion object for SystemView, giving various operations. */
object SystemView{
  type View = Views.View

  // ----- Summarising number of parameters

  /** Number of bits for each entry in the summary of the number of parameters
    * of each family in a SystemView. */
  val Shift = 6

  /** Mask to select the lowest Shift bits. */
  val Mask = (1 << Shift)-1 
  assert(Shift*numFamilies <= 32,
         "Too many families of components for current implementation.")
  assert(typeSizes.forall(size => size > 0 && size <= Mask),
        "An indexing type has too many elements for the current implementation.")

  /** Create a summary of numParams, rolling it up into a single Int. */
  @inline def mkNumParamsSummary(numParams: Array[Int]): Int = {
    var i = numTypes-1
    var numParamsSummary =  numParams(i) 
    while(i > 0){
      i -= 1
      numParamsSummary = (numParamsSummary << SystemView.Shift) + numParams(i)
    }
    numParamsSummary
  }

  // ----- Debugging

  /** Do we test various preconditions in SystemViews? */
  val Debugging = false

  /** Check the datatype invariant that should be satisfied by the components
    * of a SystemView object. */
  def checkInvariant(
    servers: List[State], components: View, numParams: Array[Int]) = {
    println("checkInvariant"); assert(false)
    // // All parameters, in the order they appear
    // val params0 = (servers++components).flatMap(_._2)
    // // All distinct non-distinguished parameters, in the order they first
    // // appear
    // val params1 = params0.filter(!isDistinguished(_)).distinct
    // for(t <- 0 until numTypes){
    //   val theseIds = params1.filter(_._1 == t).map(_._2)
    //   assert(theseIds == (0 until numParams(t)))
    // }
  }

  /** Functions that create a SystemView from its servers and components. */
  private type MkSystemView = (List[State], View) => SystemView

  /** Calculate all views of concViews of size k, with shapes in aShapes, and
    * add them to absViews.
    * @param mkSystemView a function that makes a SystemView from its servers 
    * and components. 
    * @return a list of those views that are new, i.e. not already in 
    * absViews. */
  def alpha(aShapes: List[Array[Int]], concViews: Seq[SystemView], 
            absViews: SystemViewSet, ply: Int)
      : Array[SystemView] = {
    // new views on this iteration
    var newViews = new ShardedConcBuffer[SystemView]
    for(sv <- concViews.par){
      // All views of components
      val vs = Views.alpha(aShapes, sv.componentView)
      for(v1 <- vs){ 
        val sv1 = mkSystemView(sv.servers, v1) 
        if(absViews.add(sv1)){
          newViews.put(sv1)
          assert(sv1.getPred == (null,-1))
          sv1.setPred(sv,ply)
        }
      }
    }
    newViews.get 
  }

  // ---------------- Construction

  /** Create a new SystemView from servers, components and numParams.
    *  This is the only place where SystemViews are constructed. */
  @inline private def mkSV0(
    servers: List[State], components: View, numParams: Array[Int]) = {
    // We create memory-optimised versions for sizes 4, 5, 6.
    // IMPROVE: extend to other sizes. 
    val ss = ServerStates(servers); val len = components.length
    if(len == 4)
      new SystemView4(
        ss, components(0), components(1), components(2), components(3))
    else if(len == 5)
      new SystemView5(
        ss, components(0), components(1), components(2), 
        components(3), components(4))
    else if(len == 6)
      new SystemView6(
        ss, components(0), components(1), components(2), 
        components(3), components(4), components(5))
    else new SystemViewArray(ss, components)
  }
 
  /** Factory method, creating a FullyConnectedSystemView in semi-normal 
    * form. */ 
  def mkSystemView(servers: List[State], componentView: View): SystemView = {
    // Sort componentView using insertion sort
    val len = componentView.length
    val cptsSorted = new Array[State](len); val css = new Array[Int](len)
    var i = 0
    // Inv: cptsSorted[0..i) is a sorted version of
    // componentView[0..i), and css[0..i) give the corresponding
    // control states.
    while(i < len){
      val st = componentView(i); val cs = st.cs; var j = i
      // Insert st in cptsSorted[0..i)
      while(j > 0 && css(j-1) > cs){
        cptsSorted(j) = cptsSorted(j-1); css(j) = css(j-1); j -= 1
      }
      cptsSorted(j) = st; css(j) = cs; i += 1
    }

    val (servers1, components1, args) =
      Remapper.mkCanonical2(servers, cptsSorted)
    mkSV0(servers1, components1, args)
    // new FullyConnectedSystemView(ServerStates(servers1), components1, // args,
    //                              SystemView.mkNumParamsSummary(args))
  }

  /** Factory method, creating a FullyConnectedSystemView in semi-normal
    * form. 
    * Pre: the views are sorted by control state. */ 
  def mkSystemView1(servers: List[State], componentViews: View, rhoS: ParamMap)
      : SystemView = {
    val (components1, numParams) = Remapper.mkCanonical1(componentViews, rhoS)
    mkSV0(servers, components1, numParams)
    // new FullyConnectedSystemView(
    //   ServerStates(servers), components1, // numParams, 
    //   SystemView.mkNumParamsSummary(numParams))
  }

}
