package ViewAbstraction

import scala.collection.mutable.{HashMap,HashSet}
// import ox.cads.util.Profiler

/** A set of system views. */
trait SystemViewSet{
  /** Add sv to this. */
  def add(sv: SystemView) : Boolean

  /** Add a sequence of elements to the set. */
  def add (ss: Seq[SystemView]) : Unit = for(s <- ss) add(s)

  /** Add elements of iterator to the set. */
  def add (ss: Iterator[SystemView]) : Unit = for(s <- ss) add(s)

  /** The number of elements in the set. */
  // protected var theSize = 0

  /** The number of elements in the set. */
  def size: Int // = theSize

  def reportSizes = Array(size)

  /** Iterator over the set.  Not thread safe. */
  protected def iterator : Iterator[SystemView] 

  /** List of the elements of the set.  Not thread safe. */
  def toList : List[SystemView] = iterator.toList 

  /** Make a string representing the set, separating elements using seq. */
  def mkString(sep: String) = toList.mkString(sep)

  /** Get the representative of sv.  Used in debugging.
    * Pre: this includes such a representative. */
  def getRepresentative(sv: SystemView): SystemView
}

// =======================================================

object SystemViewSet{
  def apply(): SystemViewSet = new CanonicalSystemViewSet
  //  if(useCanonical) new CanonicalSystemViewSet else new KeyedSystemViewSet
}

// =======================================================

/** An implementation based on a keyed map. */
// class KeyedSystemViewSet extends SystemViewSet{

//   /** The map storing the elements.
//     * We index the map using the key operation on SystemViews.
//     * Inv: for all cs, for all sv in underlying(cs), sv.key = cs.
//     * This represents the union of underlying(cs) for cs in dom underlying.
//     * Inv: each system view is in semi-normal form.   */
//   private val underlying = new HashMap[Key, List[SystemView]]

//   /** Add sv to this. */
//   def add(sv: SystemView) : Boolean = {
//     //val sv1 = sv.semiNorm; val cs = sv1.key
//     val key = sv.key
//     val component = underlying.getOrElseUpdate(key, List[SystemView]())
//     // Profiler.count("SystemViewSet.add"+component.size)
//     if(!component.exists(sv.equivalent(_))){
//       underlying += key -> (sv::component)
//       theSize += 1; true
//     }
//     else false
//   }

//   /** Iterator over the set. */
//   def iterator : Iterator[SystemView] = new Iterator[SystemView]{
//     private val underlyingIt = underlying.valuesIterator
//     private var list =  List[SystemView]()
//     /* This represents the Iterator with values list ++ concat(underlyingIt). */

//     /** Does the iterator have a next value? */
//     def hasNext: Boolean = list.nonEmpty || underlyingIt.hasNext

//     def next: SystemView = {
//       if(list.isEmpty) list = underlyingIt.next
//       val res = list.head; list = list.tail; res 
//     }
//   }

//   // /** Does this contain a system view equivalent to sv? */
//   // def existsEquivalent(sv: SystemView): Boolean = {
//   //   // val sv1 = sv.semiNorm;
//   //   val key = sv.key
//   //   underlying.get(key) match{
//   //     case None => false
//   //     case Some(set) =>
//   //       Profiler.count("SystemViewSet.existsEquivalent(SV)"+set.size)
//   //       set.exists(sv.equivalent)
//   //   }
//   // }

//   // def existsEquivalent(componentView: View, key: Key, viewCompanion: SystemViewT)
//   //     : Boolean = {
//   //   // Is sv equivalent to a SystemViews using componentViews, given they each
//   //   // have the same servers?
//   //   @inline def equiv(sv: SystemView) =
//   //     viewCompanion.equivalentViews(
//   //       sv.componentView, componentView, sv.rhoS)
//   //   underlying.get(key) match{
//   //     case None => false
//   //     case Some(set) => 
//   //       Profiler.count("SystemViewSet.existsEquivalent"+set.size)
//   //       set.exists(equiv)
//   //   }
//   // }

//   /** Get the representative of sv. 
//     * Pre: this includes such a representative. */
//   def getRepresentative(sv: SystemView): SystemView = {
//     val component = underlying(sv.key).toList
//     val matches = component.filter(sv.equivalent(_))
//     assert(matches.size == 1, matches)
//     matches.head
//   }
// }

// =======================================================

/** An implementation based on canonical forms. */
class CanonicalSystemViewSet extends SystemViewSet{
  /** The set of system views.  
    * Invariant: all members of underlying are in canonical form. */
  private val underlying: MyHashSet[SystemView] =
    new LockFreeReadHashSet[SystemView] // (powerOf2AboveNumThreads*8)
    // new MyShardedHashSet[SystemView]

  /** Add sv to this. */
  @inline def add(sv: SystemView): Boolean = underlying.add(sv) 
  //  if(underlying.add(sv)){ theSize += 1; true } else false

  override def size = underlying.size

  override def reportSizes = underlying.reportSizes

  /** Iterator over the set. */
  protected def iterator: Iterator[SystemView] =
    underlying.iterator 

  /** Get the representative of sv.  Used in debugging.
    * Pre: this includes such a representative. */
  def getRepresentative(sv: SystemView): SystemView =  underlying.get(sv)
    // In the implementation that uses this, only canonical forms are used.
    // Check this is in canonical form, and in underlying.
    // val (cpts1, _) = Remapper.mkCanonical1(sv.componentView, sv.rhoS)
    // assert(cpts1 == sv.componentView)
    // assert(underlying.contains(sv), sv)
    // sv

}
