package ViewAbstraction

import scala.collection.mutable.Set
// import ox.cads.util.Profiler
import scala.math.Ordering.Implicits._ // for sorting list of lists


/** Various operations on views of ring processes. */
object Views{
  /** Views of a family of replicated processes. */
  type View = Array[State]

  assert(numTypes >= 0 && maxSubtypeSize >= 0 && numTypes*maxSubtypeSize <= 64,
         "Too many identities in script")

  def show(v: View): String = v.mkString("<", ", ", ">")

  /** A list of identitites, summarised as a bitmap.  This version depends upon
    * the assumption captured by the above assertion. */
  // def idsBitMap(ids: List[Parameter]): Long = {
  //   // The bit to use for (t,i)
  //   def bitFor(t: Int, i: Int) = t*maxSubtypeSize+i
  //   var bitmap = 0L
  //   for((t,i) <- ids) bitmap |= 1 << bitFor(t,i)
  //   bitmap
  // }

  /** Insert s into view in place of its ith member (indexing from 0). */
  def insert(view: View, i: Int, s: State): View = {
    // view.patch(i, List(s), 1)
    // Array.tabulate(view.length)(j =>
    //   if(j < i) view(j) else if(j == i) s else view(j))
    val v1 = view.clone; v1(i) = s; v1
  }

  /** Insert each state in states into view in place of its ith member (indexing
    * from 0). */
  @inline def insertStates(view: View, i: Int, states: List[State]): List[View]
    = {
    states.map(insert(view, i, _))
    // val (before, _::after) = view.splitAt(i)
    // states.map(s => before++(s::after))
  }

  /** Insert each state of states0 into view in position i0, and insert each
    * state of states1 in position i1. */
  @inline def insert2States(view: View, i0: Int, states0: List[State],
                            i1: Int, states1: List[State]): List[View] = {
    for(s0 <- states0; s1 <- states1) yield{
      val v1 = view.clone; v1(i0) = s0; v1(i1) = s1; v1
    }
    // if(i0 < i1){
    //   val (before1, _::after) = view.splitAt(i1)
    //   val (before, _::mid) = before1.splitAt(i0)
    //   val tails = states1.map(s1 => mid++(s1::after))
    //   for(tail <- tails; s0 <- states0) yield before++(s0::tail)
    // }
    // else insert2States(view, i1, states1, i0, states0)
  }

  /** The shape of a view, i.e. an array giving the number of states of each
    * process type. */
  def shape(view: View): Array[Int] = {
    val theShape = new Array[Int](numFamilies)
    for(st <- view) theShape(st.family) += 1
    theShape
  }

  /** Add view to views if it is not equal to any existing member. */
  // @inline private def maybeAdd(views: List[View], view: View) : List[View] = {
  //   // assert(!views.contains(view)); 
  //   view::views 
  // }

  /** All views of view with shapes from aShapes. */
  def alpha(aShapes: List[Array[Int]], view: View) : List[View] = {
    // Restrictions of view to each family
    // println("alpha: "+show(view))
    val subViews = Array.tabulate(numFamilies)(f =>
      view.filter(_.family == f) ) //State.idType(st.cs) == t) )
    // println("subViews = "+subViews.map(show).mkString("; "))
    // val subViews = Array.tabulate(numTypes)(t =>
    //   view.filter(_.idType == t) ) //State.idType(st.cs) == t) )
    val lengths = subViews.map(_.length)
    // For each family f, all views of the restriction of view to f.
    val allAlphas = Array.tabulate(numFamilies)(f =>
      alpha2(lengths(f), subViews(f)) )
    // val allAlphas = Array.tabulate(numTypes)(t =>
    //   alpha2(lengths(t), subViews(t)) )
    for(sh <- aShapes; if(0 until numFamilies).forall(f => sh(f) <= lengths(f));
    // for(sh <- aShapes; if(0 until numTypes).forall(t => sh(t) <= lengths(t));
        // Create all subviews of size sh(t) for each type t, and merge them
        // in all possible ways.
        v <- allMerges(Array.tabulate(numFamilies)(f => allAlphas(f)(sh(f))))
        // v <- allMerges(Array.tabulate(numTypes)(t => allAlphas(t)(sh(t))))
    )
    yield v
  }

  /** All views of view[start..) of sizes up to k, in an array, indexed
    * by size. */
  private def alpha2(k: Int, view: View) : Array[List[View]] = {
    val views = Array.fill(k+1)(List[View]())
    views(0) = List(Array[State]())
    // Inv: for all j, views(j) gives all views of size j of
    // view[len-i..len); so if i<j then views(i) is empty.
    val len = view.length; var i = 0
    while(i < len){
      i += 1
      val s0 = view(len-i)
      // for(j <- i to k) assert(views(j).isEmpty)
      // Add s0 to each view of size j-1 to create a view of size j
      var j = k min i
      while(j >= 1){
        for(v <- views(j-1)) views(j) ::= s0+:v
        j -= 1
      }
    }
    // for(i <- 0 until k+1)  println(i+": "+views(i).map(show))
    views
  }

  /** All ways of merging one element from each element of subViews. */
  private def allMerges(subViews: Array[List[View]]): List[View] = {
    // println("allMerges: "+subViews.map(_.map(show)).mkString("; "))
    // All ways of merging one element from each element of
    // subViews[f..numtypes). */
    def allMergesFrom(f: Int): List[View] =
      if(f == numFamilies-1) subViews(f)
      else{
        val recMerges = allMergesFrom(f+1)
        for(v1 <- subViews(f); v2 <- recMerges) yield merge(v1, v2)
      }
    allMergesFrom(0)
  }

   /** Merge views v1 and v2, giving a view with control states in order.  
    * 
    * Pre: the two views have distinct control states; and each has control
    * states in order. */
  private def merge(v1: View, v2: View): View = {
    val len1 = v1.length; val len2 = v2.length
    val result = new Array[State](len1+len2)
    var i1 = 0; var i2 = 0; var i = 0 // Inv i = i1+i2
    while(i1 < len1 && i2 < len2){
      val st1 = v1(i1); val st2 = v2(i2)
      if(st1.cs < st2.cs){ result(i) = st1; i1 += 1; i += 1 }
      else{ assert(st2.cs < st1.cs); result(i) = st2; i2 += 1; i += 1 }
    }
    while(i1 < len1){ result(i) = v1(i1); i1 += 1; i += 1 }
    while(i2 < len2){ result(i) = v2(i2); i2 += 1; i += 1 }
    result
  }
    // if(v1.isEmpty) v2
    // else if(v2.isEmpty) v1
    // else{
    //   val st1 = v1.head; val st2 = v2.head
    //   if(st1.cs < st2.cs) st1 :: merge(v1.tail, v2)
    //   else{ assert(st2.cs < st1.cs); st2 :: merge(v1, v2.tail) }
    // }

}
